# import pandas as pd
import os
import math

# import argparse
# import datetime
import logging
import time

from spacemake.config import Puck
from spacemake.errors import *

from spacemake.config import ConfigFile
from spacemake.util import message_aggregation, assert_file, str_to_list
from spacemake.snakemake.variables import puck_barcode_files_summary
from typing import List, Dict

logger_name = "spacemake.project_df"
logger = logging.getLogger(logger_name)


class ProjectDF:
    """
    ProjectDF: class responsible for managing spacemake projects.

    :param file_path: path to the project_df.csv file, where we will save the project list.
    :param config: config file object
    :type config: ConfigFile
    :param df: A pandas dataframe, containing one row per sample
    :type df: pd.DataFrame
    """

    logger = logging.getLogger("spacemake.project_df.ProjectDF")

    # default values of the project dataframe columns
    project_df_default_values = {
        "puck_barcode_file_id": ["no_spatial_data"],
        "sample_sheet": None,
        "species": None,
        "demux_barcode_mismatch": 1,
        "demux_dir": None,
        "basecalls_dir": None,
        "R1": None,
        "R2": None,
        "reads": None,
        "longreads": None,
        "longread_signature": None,
        "investigator": "unknown",
        "sequencing_date": "unknown",
        "experiment": "unknown",
        "puck_barcode_file": None,
        "run_mode": ["default"],
        "barcode_flavor": "default",
        "is_merged": False,
        "merged_from": [],
        "puck": "default",
        "dge": None,
        "map_strategy": {  # default mapping strategy changes depending on if we have rRNA or not
            True: "rRNA:bowtie2->genome:STAR",  # map in parallel to rRNA and genome (default so far)
            False: "genome:STAR",  # w/o rRNA, map to genome directly
        },
        "adapter_flavor": "default",
    }

    project_df_dtypes = {
        "puck_barcode_file_id": "object",
        "sample_sheet": "str",
        "species": "str",
        "demux_barcode_mismatch": "int64",
        "demux_dir": "str",
        "basecalls_dir": "str",
        "R1": "object",
        "R2": "object",
        "reads": "object",
        "longreads": "str",
        "longread_signature": "str",
        "investigator": "str",
        "sequencing_date": "str",
        "experiment": "str",
        "puck_barcode_file": "object",
        "run_mode": "object",
        "barcode_flavor": "str",
        "is_merged": "bool",
        "merged_from": "object",
        "puck": "str",
        "dge": "str",
    }

    def __init__(self, file_path, config: ConfigFile = None):
        """__init__.

        :param file_path: path to pandas data frame (saved as .csv)
        :param config: ConfigFile
        :type config: ConfigFile
        """
        self.file_path = file_path
        self.config = config
        import pandas as pd

        if os.path.isfile(file_path):
            attempts = 0
            df_loaded_flag = False

            while not df_loaded_flag:
                try:
                    self.df = pd.read_csv(
                        file_path,
                        index_col=["project_id", "sample_id"],
                        na_values=["None", "none"],
                        dtype=self.project_df_dtypes,
                    )
                    df_loaded_flag = True
                except pd.errors.EmptyDataError as e:
                    if attempts < 5:
                        # wait 5 seconds before retrying
                        self.logger.warning(
                            f"The file '{self.file_path}' seems to be empty. Trying again ({attempts}/5) in 5 seconds"
                        )
                        time.sleep(5)
                        attempts = attempts + 1
                        continue
                    else:
                        raise e

            if self.df.empty:
                self.create_empty_df()
            else:
                # 'fix' the dataframe if there are inconsistencies
                self.fix()
        else:
            self.create_empty_df()

        self.assert_valid()

        self.logger = logging.getLogger(logger_name)

    def assert_valid(self):
        """assert_valid.

        this function iterates over projects/samples in the project_df, and asserts
        whether the specified variables are in accordance with the configuration file,
        and whether the specified files (R1, R2, puck_barcode_files) exist at the
        specified locations.
        """

        project_df_column_to_config_varname = {
            "run_mode": "run_modes",
            "species": "species",
            "barcode_flavor": "barcode_flavors",
            "puck": "pucks",
        }
        if not hasattr(self, "df"):
            raise SystemExit(
                ValueError(
                    "The 'project_df' does not exist in the ProjectDF object or is empty"
                )
            )
        elif self.df.empty:
            logger.warning("The 'project_df' in the ProjectDF object is empty")
            self.create_empty_df()

        _unique_samples = self._check_unique_samples()
        if _unique_samples is not None:
            raise SampleAlreadyExistsError(_unique_samples)

        for index, row in self.df.iterrows():
            # check that sample sheet file exists
            if (row["sample_sheet"] is not None) and (
                not os.path.exists(row["sample_sheet"])
            ):
                raise SystemExit(
                    FileNotFoundError(
                        f"At {index}, the 'sample_sheet' file does not exist"
                    )
                )

            # checking that variables are what they're supposed to be (according to config file)
            for pdf_col, config_var in project_df_column_to_config_varname.items():
                if type(row[pdf_col]) is list:
                    for _it_row in row[pdf_col]:
                        self.config.assert_variable(f"{config_var}", _it_row)
                elif type(row[pdf_col]) is str:
                    self.config.assert_variable(f"{config_var}", row[pdf_col])

            # check that puck_barcode_file(s), R1 and R2 files exist
            for n_col in ["puck_barcode_file", "R1", "R2"]:
                if n_col in ["R1", "R2"] and row["is_merged"]:
                    continue

                if type(row[n_col]) is list:
                    for _it_row in row[n_col]:
                        if not os.path.exists(_it_row):
                            raise SystemExit(
                                FileNotFoundError(
                                    f"At {index}, the {n_col} file does not exist: '{os.path.abspath(_it_row)}'"
                                )
                            )
                elif type(row[n_col]) is str and row[n_col] != "":
                    if not os.path.exists(row[n_col]):
                        raise SystemExit(
                            FileNotFoundError(
                                f"At {index}, the {n_col} file does not exist: '{os.path.abspath(_it_row)}'"
                            )
                        )

            # check that pucks are specified only if puck is spatial (or puck_collection is enabled)
            _valid_puck_coordinate = True
            if row["puck_barcode_file"] is None:
                _valid_puck_coordinate = False
            elif type(row["puck_barcode_file"]) is list:
                if len(row["puck_barcode_file"]) == 0:
                    _valid_puck_coordinate = False
            elif (
                type(row["puck_barcode_file"]) is str and row["puck_barcode_file"] == ""
            ):
                _valid_puck_coordinate = False

            # check that merging variables are properly specified
            # otherwise, this throws 'MissingInputException' because it cannot find the bam files
            if row["is_merged"]:
                if len(row["merged_from"]) < 2:
                    raise SystemExit(
                        SpacemakeError(
                            f"At {index}, there is <2 samples under 'merged_from'"
                        )
                    )
                for merged_i in row["merged_from"]:
                    if (
                        not isinstance(merged_i, tuple)
                        and not isinstance(merged_i, list)
                    ) or len(merged_i) != 2:
                        raise SystemExit(
                            SpacemakeError(
                                f"At {index}, wrong format for 'merged_from'.\n"
                                + "It must be something like "
                                + "[('project', 'sample_a'), ('project', 'sample_b')]"
                            )
                        )
                    try:
                        self.assert_sample(merged_i[0], merged_i[1])
                    except ProjectSampleNotFoundError as e:
                        self.logger.error(f"Merging error at {index}")
                        raise e

            if not _valid_puck_coordinate:
                _puck_vars = self.get_puck_variables(
                    project_id=index[0], sample_id=index[1]
                )
                if _puck_vars.get("coordinate_system", "") != "":
                    # TODO: This can arise if no tiles have sufficient overlap
                    # instead of repeatedly complaining, we should
                    # * issue a big fat warning!
                    # * replace 'puck_barcode_file_id' with ['no_spatial_data'] and
                    # * set 'puck' to empty?? (I would like to keep the warning in subsequent
                    # runs, rather than pretending that this always was 'no_spatial_data')
                    self.logger.warning(
                        f"At {index}, the selected puck '{row['puck']}' "
                        "contains a coordinate_system "
                        "but no 'puck_barcode_files' are specified. "
                        "no spatial information will be available."
                    )
                    self.df.at[index, "puck_barcode_file"] = None
                    self.df.at[index, "puck_barcode_file_id"] = []

            # print(self.df[["puck_barcode_file_id", "puck_barcode_file", "puck"]])

    def _check_unique_samples(self):
        for index, _ in self.df.iterrows():
            if len(self.df[self.df.index.isin([index])]) > 1:
                return index

        return None

    def create_empty_df(self):
        import pandas as pd

        index = pd.MultiIndex(
            names=["project_id", "sample_id"], levels=[[], []], codes=[[], []]
        )
        self.df = pd.DataFrame(self.project_df_default_values, index=index)

        self.df = self.df.astype(self.project_df_dtypes)
        self.dump()

    def compute_max_barcode_mismatch(self, indices: List[str]) -> int:
        """compute_max_barcode_mismatch.

        :param indices: List of illumina I7 index barcodes
        :type indices: List[str]
        :return: the maximum mismatch to be allowed for this set of index barcodes
        :rtype: int
        """
        num_samples = len(indices)

        if num_samples == 1:
            return 4
        else:
            max_mismatch = 3
            for i in range(num_samples - 1):
                for j in range(i + 1, num_samples):
                    hd = self.hamming_distance(indices[i], indices[j])
                    max_mismatch = min(max_mismatch, math.ceil(hd / 2) - 1)

        return max_mismatch

    def hamming_distance(self, string1: str, string2: str) -> int:
        """Cacluate hamming distance between two strings

        :param string1:
        :type string1: str
        :param string2:
        :type string2: str
        :rtype: int
        """
        return sum(c1 != c2 for c1, c2 in zip(string1, string2))

    def find_barcode_file(self, puck_barcode_file_id: str) -> str:
        """Tries to find path of a barcode file, using the puck_barcode_file_id.

        :param puck_barcode_file_id: puck_barcode_file_id of the puck we are looking for.
        :type puck_barcode_file_id: str
        :return: path of the puck file, containing barcodes, or None
        :rtype: str
        """
        # first find directory of puck file

        # return none or the path of the file
        def get_barcode_file(path):
            """get_barcode_file.

            :param path:
            """
            if os.path.isfile(path):
                return path

            return None

        def find_dir(name, path):
            """find_dir.

            :param name:
            :param path:
            """
            for root, dirs, files in os.walk(path):
                if name in dirs:
                    return os.path.join(root, name)

        puck_dir = find_dir(puck_barcode_file_id, self.config.puck_data["root"])
        path = None

        if puck_dir is not None:
            # puck dir exists, look for barcode file pattern
            path = os.path.join(puck_dir, self.config.puck_data["barcode_file"])

            return get_barcode_file(path)
        else:
            return self.project_df_default_values["puck_barcode_file"]

    def consolidate_pucks_merged_samples(self):
        for index, row in self.df.iterrows():
            project_id, sample_id = index
            puck_ids = row["puck_barcode_file_id"]

            if (not row["is_merged"]) or (
                not self.is_spatial(project_id, sample_id, puck_ids)
            ):
                continue

            if len(puck_ids) >= 1:
                puck_ids = puck_ids[0]
            elif len(puck_ids):
                puck_ids = self.project_df_default_values["puck_barcode_file_id"]

            merged_from = row["merged_from"]
            puck_id_file = set()

            for sample_tuple in merged_from:
                pid = self.df.loc[sample_tuple]["puck_barcode_file_id"]
                pbf = self.df.loc[sample_tuple]["puck_barcode_file"]
                _tuple = [(id, bf) for id, bf in zip(pid, pbf)]

                puck_id_file.update([tuple(t) for t in _tuple])

            pid, pbf = list(zip(*list(puck_id_file)))
            self.df.loc[index, "puck_barcode_file_id"] = list(pid)
            self.df.loc[index, "puck_barcode_file"] = list(pbf)

    def update_project_df_barcode_matches(self, prealigned=False):
        from spacemake.snakemake.variables import (
            puck_count_barcode_matches_summary,
            puck_count_prealigned_barcode_matches_summary,
        )
        import pandas as pd

        if prealigned:
            _bc_file = puck_count_prealigned_barcode_matches_summary
        else:
            _bc_file = puck_count_barcode_matches_summary

        for index, row in self.df.iterrows():
            project_id, sample_id = index

            puck_ids = row["puck_barcode_file_id"]
            if len(puck_ids) >= 1:
                puck_ids = puck_ids[0]
            elif len(puck_ids):
                puck_ids = self.project_df_default_values["puck_barcode_file_id"]

            if (row["is_merged"] and prealigned) or (
                not self.is_spatial(project_id, sample_id, puck_ids)
            ):
                continue

            # check if barcodes have been filtered
            _f_barcodes_df = _bc_file.format(project_id=project_id, sample_id=sample_id)

            # project_df is only updated if a prealignment barcode matching file is found
            if os.path.exists(_f_barcodes_df):
                barcodes_df = pd.read_csv(_f_barcodes_df)

                if "pass_threshold" not in barcodes_df.columns:
                    continue

                above_threshold_mask = barcodes_df["pass_threshold"] == 1

                _puck_barcode_files = barcodes_df[above_threshold_mask][
                    "puck_barcode_file"
                ].values.tolist()
                _puck_barcode_files_id = barcodes_df[above_threshold_mask][
                    "puck_barcode_file_id"
                ].values.tolist()

                self.df.at[index, "puck_barcode_file"] = _puck_barcode_files
                self.df.at[index, "puck_barcode_file_id"] = _puck_barcode_files_id

    def get_sample_info(self, project_id: str, sample_id: str) -> Dict:
        """get_sample_info.

        :param project_id:
        :type project_id: str
        :param sample_id:
        :type sample_id: str
        :return: A dictionary containing all the values of a given sample, from the ProjectDF.
        :rtype: Dict
        """
        # returns sample info from the projects df
        self.assert_sample(project_id, sample_id)
        out_dict = self.df.loc[(project_id, sample_id)].to_dict()

        return out_dict

    def is_external(self, project_id: str, sample_id: str) -> bool:
        """is_external.

        :param project_id:
        :type project_id: str
        :param sample_id:
        :type sample_id: str
        :rtype: bool
        """
        self.assert_sample(project_id, sample_id)

        data = self.df.loc[
            (project_id, sample_id),
            [
                "R1",
                "R2",
                "reads",
                "basecalls_dir",
                "sample_sheet",
                "longreads",
                "dge",
                "is_merged",
            ],
        ]

        self.logger.debug(
            f"project_id={project_id} sample_id={sample_id} "
            + f"R1,2={bool(data.R1 and data.R2)} "
            + f"basecall={(data.basecalls_dir and data.sample_sheet)} "
            + f"longreads={bool(data.longreads)} dge={bool(data.dge)} "
            + f"is_merged={bool(data.is_merged)}"
        )
        if (
            data.R2  # R1 is optional (bulk samples)
            or (data.basecalls_dir and data.sample_sheet)
            or (data.longreads)
            or (data.reads)
            and not data.dge
            or data.is_merged
        ):
            return False
        elif data.dge:
            return True
        else:
            raise SpacemakeError(
                f"Sample with id (project_id, sample_id)="
                + f"({project_id}, {sample_id}) is invalid."
            )

    def has_dge(self, project_id: str, sample_id: str) -> bool:
        """Returns True if a has dge. for Pacbio only samples returns False.

        :param project_id:
        :type project_id: str
        :param sample_id:
        :type sample_id: str
        :rtype: bool
        """
        self.assert_sample(project_id, sample_id)

        data = self.df.loc[
            (project_id, sample_id),
            [
                "R1",
                "R2",
                "reads",
                "basecalls_dir",
                "sample_sheet",
                "longreads",
                "dge",
                "is_merged",
            ],
        ]

        if (
            data.is_merged
            or data.R2
            or (data.sample_sheet and data.basecalls_dir)
            or data.dge
            or data.reads
        ):
            return True
        elif data.longreads:
            return False
        else:
            raise SpacemakeError(
                f"Sample with id (project_id, sample_id)="
                + f"({project_id}, {sample_id}) is invalid."
            )

    def is_spatial(
        self, project_id: str, sample_id: str, puck_barcode_file_id: str
    ) -> bool:
        """Returns true if a sample with index (project_id, sample_id) is spatial,
        meaning that it has spatial barcodes attached. Or, if the puck_barcode_file_id
        is 'puck_collection', meaning that the same is necessarily spatial (transformed
        from local to global coordinates)

        :param project_id:
        :type project_id: str
        :param sample_id:
        :type sample_id: str
        :rtype: bool
        """
        self.assert_sample(project_id, sample_id)
        puck_barcode_file = self.get_puck_barcode_file(
            project_id=project_id,
            sample_id=sample_id,
            puck_barcode_file_id=puck_barcode_file_id,
        )

        if puck_barcode_file is not None or puck_barcode_file_id == "puck_collection":
            return True
        else:
            return False

    def get_default_map_strategy_for_species(self, species):
        have_rRNA = "rRNA" in self.config.variables["species"][species]
        map_strategy = self.project_df_default_values["map_strategy"][have_rRNA]
        # print(
        #     f"getting default for '{species}' have_rRNA={have_rRNA} -> {map_strategy}"
        # )
        return map_strategy

    def fix(self):
        import numpy as np
        import pandas as pd

        modified = False
        # convert types
        self.df = self.df.where(pd.notnull(self.df), None)
        self.df = self.df.replace({np.nan: None})
        # replacing NaN with None

        # rename puck_id to puck_barcode_file_id, for backward
        # compatibility
        self.df.rename(
            columns={"puck_id": "puck_barcode_file_id"},
            inplace=True,
        )

        # convert list values stored as string
        self.df.run_mode = self.df.run_mode.apply(str_to_list)
        self.df.merged_from = self.df.merged_from.apply(str_to_list)

        # convert R1/R2 to list, if they are stored as string
        self.df.R1 = self.df.R1.apply(str_to_list)
        self.df.R2 = self.df.R2.apply(str_to_list)

        self.df.puck_barcode_file_id = self.df.puck_barcode_file_id.apply(str_to_list)
        self.df.puck_barcode_file = self.df.puck_barcode_file.apply(str_to_list)

        project_list = []
        # required if upgrading from pre-longread tree
        if not "reads" in self.df.columns:
            modified = True
            self.df["reads"] = None

        if not "longreads" in self.df.columns:
            modified = True
            self.df["longreads"] = None

        if not "longread_signature" in self.df.columns:
            modified = True
            self.df["longread_signature"] = None

        # required if upgrading from pre-bowtie2/map-strategy tree
        if not "map_strategy" in self.df.columns:
            modified = True
            self.df["map_strategy"] = self.df["species"].apply(
                self.get_default_map_strategy_for_species
            )

        # validate and correct map_strategies
        from spacemake.map_strategy import validate_mapstr

        corrected_map_strategies = []
        for row in self.df.itertuples():
            corrected_map_strategies.append(
                validate_mapstr(
                    row.map_strategy, config=self.config, species=row.species
                )
            )
        self.df["map_strategy"] = corrected_map_strategies

        if not "adapter_flavor" in self.df.columns:
            self.df["adapter_flavor"] = self.project_df_default_values["adapter_flavor"]
            # print("added adapter-flavor!")
            modified = True

        if modified:
            self.logger.warning(
                f".fix() reported changes! Saving migrated project_df.csv to '{self.file_path}'"
            )
            # self.logger.warning(self.df)
            # self.logger.warning(self.df.columns)
            # self.logger.warning(self.df["adapter_flavor"])
            self.dump()

        # per row updates
        # first create a series of a
        for ix, row in self.df.iterrows():
            s = pd.Series(self.project_df_default_values)

            # update puck barcode file info
            # for samples which have shared barcodes, and this barcode info is
            # stored in a puck, in the config file, before the id was set to
            # the name of the puck, and the puck_barcode_file was set to None.
            # Here we populate the puck_barcode_file into the path to the actual
            # file so that no errors are caused downstream.
            if (
                row["puck_barcode_file"] is None
                and row["puck_barcode_file_id"] is not None
            ):
                if len(row["puck_barcode_file_id"]) > 1:
                    raise SpacemakeError(
                        "When no barcode file provided, there "
                        + "only should be one id available"
                    )

                elif len(row["puck_barcode_file_id"]) < 1:
                    self.logger.warning(
                        "empty puck_barcode_file_id list detected. "
                        "Perhaps something went wrong intersecting your samples barcodes with your sample? "
                        "No spatial data will be available."
                    )
                    pbf_id = "no_spatial_data"
                else:
                    pbf_id = row["puck_barcode_file_id"][0]

                if pbf_id not in self.project_df_default_values["puck_barcode_file_id"]:
                    puck = self.config.get_puck(pbf_id)

                    row["puck_barcode_file"] = [puck.variables["barcodes"]]

            s.update(row)
            s.name = row.name
            project_list.append(s)

        self.df = pd.concat(project_list, axis=1).T
        self.df.is_merged = self.df.is_merged.astype(bool)
        self.df.index.names = ["project_id", "sample_id"]

    def get_puck_barcode_file(
        self, project_id: str, sample_id: str, puck_barcode_file_id: str
    ) -> str:
        if (
            puck_barcode_file_id
            in self.project_df_default_values["puck_barcode_file_id"]
        ):
            # if sample is not spatial, or we request the non-spatial puck
            return None
        else:
            ids = self.get_metadata(
                "puck_barcode_file_id", sample_id=sample_id, project_id=project_id
            )

            puck_barcode_files = self.get_metadata(
                "puck_barcode_file", sample_id=sample_id, project_id=project_id
            )

            # if no puck_barcode_file is provided, it means that barcode
            # file has to be fetched from the puck itself
            if puck_barcode_files is not None:
                for pid, pbf in zip(ids, puck_barcode_files):
                    if pid == puck_barcode_file_id:
                        return pbf

                return None

    def get_puck_barcode_ids_and_files(
        self,
        project_id: str,
        sample_id: str,
    ) -> str:
        puck_barcode_file_ids = self.get_metadata(
            "puck_barcode_file_id", project_id=project_id, sample_id=sample_id
        )

        puck_barcode_files = self.get_metadata(
            "puck_barcode_file", project_id=project_id, sample_id=sample_id
        )

        out_puck_barcode_files = []
        out_puck_barcode_file_ids = []

        # return only id-file pairs, for which file is not none
        if puck_barcode_files is not None:
            for pbf_id, pbf in zip(puck_barcode_file_ids, puck_barcode_files):
                out_puck_barcode_files.append(pbf)
                out_puck_barcode_file_ids.append(pbf_id)

        return out_puck_barcode_file_ids, out_puck_barcode_files

    def get_matching_puck_barcode_file_ids(
        self,
        project_id: str,
        sample_id: str,
    ):
        import pandas as pd

        summary_file = puck_barcode_files_summary.format(
            project_id=project_id, sample_id=sample_id
        )

        if not os.path.isfile(summary_file):
            # print(f"looking for summary file: '{summary_file}'")
            return self.project_df_default_values["puck_barcode_file_id"]

        df = pd.read_csv(summary_file)

        # Comment the following line that restricts the analysis of some tiles.
        # All tiles provided will now be processed -- it's up to the user to
        # provide a meaningful list of tiles.
        #
        # df = df.loc[(df.n_matching > 500) & (df.matching_ratio > 0.1)]

        pdf_ids = df.puck_barcode_file_id.to_list()

        pdf_ids.append(self.project_df_default_values["puck_barcode_file_id"][0])

        return pdf_ids

    def get_puck_barcode_file_metrics(
        self,
        project_id: str,
        sample_id: str,
        puck_barcode_file_id: str,
    ):
        import numpy as np

        summary_file = puck_barcode_files_summary.format(
            project_id=project_id, sample_id=sample_id
        )

        if not os.path.isfile(summary_file):
            return None
        import pandas as pd

        df = pd.read_csv(summary_file)

        df_puck = df.loc[df.puck_barcode_file_id == puck_barcode_file_id]

        # we calculate the stats for the puck_collection here
        # the max and min coordinates are not in global system
        if puck_barcode_file_id == "puck_collection":

            def multi_func(functions):
                def f(col):
                    return functions[col.name](col)

                return f

            df_pc = df._get_numeric_data().apply(
                multi_func(
                    {
                        "x_pos_min_px": np.min,
                        "x_pos_max_px": np.max,
                        "y_pos_min_px": np.min,
                        "y_pos_max_px": np.max,
                        "n_barcodes": np.mean,
                        "n_matching": np.mean,
                        "matching_ratio": np.mean,
                        "px_by_um": np.mean,
                    }
                )
            )

            return df_pc.to_dict()

        if df_puck.empty:
            return None
        else:
            return df_puck.iloc[0].to_dict()

    def get_puck(self, project_id: str, sample_id: str, return_empty=False) -> Puck:
        """get_puck.

        :param project_id: project_id of a sample
        :type project_id: str
        :param sample_id: sample_id of a sample
        :type sample_id: str
        :param return_empty:
        :return: A Puck object containing puck object
        :rtype: Puck
        """
        puck_name = self.get_metadata(
            "puck", project_id=project_id, sample_id=sample_id
        )

        return self.config.get_puck(puck_name, return_empty=return_empty)

    def get_puck_variables(
        self, project_id: str, sample_id: str, return_empty=False
    ) -> Dict:
        """get_puck_variables.

        :param project_id: project_id of a sample
        :type project_id: str
        :param sample_id: sample_id of a sample
        :type sample_id: str
        :param return_empty:
        :return: A dictionary containing the puck variables of a given sample
        :rtype: Dict
        """
        puck_name = self.get_metadata(
            "puck", project_id=project_id, sample_id=sample_id
        )

        return self.config.get_puck(puck_name, return_empty=return_empty).variables

    def get_metadata(self, field, project_id=None, sample_id=None, **kwargs):
        """get_metadata.

        :param field:
        :param project_id:
        :param sample_id:
        :param kwargs:
        """
        df = self.df
        if sample_id is not None:
            df = df.query("sample_id == @sample_id")

        if project_id is not None:
            df = df.query("project_id == @project_id")

        for key, value in kwargs.items():
            df = df.loc[df.loc[:, key] == value]

        dl = df.loc[:, field].to_list()
        if len(dl):
            return dl[0]
        else:
            return ""

    def dump(self):
        """dump."""
        self.logger.debug(f"writing project_df.csv to '{self.file_path}")
        self.df.to_csv(self.file_path)

    def add_sample_sheet(self, sample_sheet_path, basecalls_dir):
        """add_sample_sheet.

        :param sample_sheet_path:
        :param basecalls_dir:
        """
        import pandas as pd

        with open(sample_sheet_path) as sample_sheet:
            ix = 0
            investigator = None
            sequencing_date = None

            for line in sample_sheet:
                line = line.strip("\n")
                if "Investigator" in line:
                    investigator = line.split(",")[1]
                if "Date" in line:
                    sequencing_date = line.split(",")[1]
                if "[Data]" in line:
                    # the counter ix stops here
                    break
                else:
                    ix = ix + 1

        # read everything after [Data]
        df = pd.read_csv(sample_sheet_path, skiprows=ix + 1)
        # rename columns
        to_rename = {
            "Sample_ID": "sample_id",
            "Sample_Name": "puck_barcode_file_id",
            "Sample_Project": "project_id",
            "Description": "experiment",
            "index": "index",
        }
        df.rename(
            columns=to_rename,
            inplace=True,
        )
        # select only renamed columns
        df = df[to_rename.values()]
        df["species"] = df["experiment"].str.split("_").str[-1]
        df["investigator"] = investigator
        df["sequencing_date"] = sequencing_date

        # rename columns
        df["basecalls_dir"] = basecalls_dir
        df["demux_barcode_mismatch"] = self.compute_max_barcode_mismatch(df["index"])
        df["sample_sheet"] = sample_sheet_path
        df["demux_dir"] = (
            df["sample_sheet"].str.split("/").str[-1].str.split(".").str[0]
        )
        df["puck_barcode_file"] = df.puck_barcode_file_id.apply(self.find_barcode_file)
        df.set_index(["project_id", "sample_id"], inplace=True)

        for ix, row in df.iterrows():
            self.add_update_sample(project_id=ix[0], sample_id=ix[1], **row.to_dict())

    def assert_index_value(self, index_value, index_level):
        if not isinstance(index_value, list):
            index_value = [index_value]

        ixs = self.df.index.get_level_values(index_level)

        for ixv in index_value:
            if ixv not in ixs:
                raise ProjectSampleNotFoundError(index_level, ixv)

    def sample_exists(self, project_id=None, sample_id=None):
        """sample_exists.

        :param project_id:
        :param sample_id:
        """
        if project_id is None or sample_id is None:
            raise Exception(
                f"you need to provide a sample_id and project_id to check if sample exists"
            )
        else:
            ix = (project_id, sample_id)

            return ix in self.df.index

    def assert_sample(self, project_id, sample_id):
        if not self.sample_exists(project_id, sample_id):
            raise ProjectSampleNotFoundError(
                "(project_id, sample_id)", (project_id, sample_id)
            )

    def assert_run_mode(self, project_id, sample_id, run_mode_name):
        variables = self.get_sample_info(project_id, sample_id)

        if run_mode_name not in variables["run_mode"]:
            raise SpacemakeError(
                f"(project_id, sample_id)=({project_id},"
                + f"{sample_id}) has no run_mode={run_mode_name}\n"
                + f'run_mode has to be one of {variables["run_mode"]}'
            )

    def add_update_sample(
        self,
        action=None,
        project_id=None,
        sample_id=None,
        R1=None,
        R2=None,
        reads=None,
        dge=None,
        longreads=None,
        longread_signature=None,
        sample_sheet=None,
        basecalls_dir=None,
        is_merged=False,
        return_series=False,
        map_strategy=None,
        puck_barcode_file=None,
        puck_barcode_file_id=None,
        **kwargs,
    ):
        """add_update_sample.

        :param action:
        :param project_id:
        :param sample_id:
        :param R1:
        :param R2:
        :param reads:
        :param dge:
        :param longreads:
        :param longread_signature:
        :param sample_sheet:
        :param basecalls_dir:
        :param is_merged:
        :param return_series:
        :param map_strategy:
        :param kwargs:
        """
        import pandas as pd

        ix = (project_id, sample_id)
        sample_exists = self.sample_exists(*ix)

        if action is None and sample_exists:
            action = "update"
        elif action is None and not sample_exists:
            action = "add"
        elif action == "add" and sample_exists:
            raise SampleAlreadyExistsError(ix)
        elif action == "update" and not sample_exists:
            raise ProjectSampleNotFoundError("(project_id, sample_id)", ix)

        if action == "add":
            self.logger.info(f"Adding sample {ix}")
        elif action == "update":
            self.logger.info(f"Updating sample {ix}")
        else:
            raise ValueError(f"Unknown action {action}")

        # check variables
        # First we check if R1 and R2 is present.
        # If not, we check for longreads. If provided, we also need
        #   --longread-signature
        # If longreads not provided, we try with basecalls_dir and sample_sheet
        #   (only used by add_sample_sheet command)
        # If those area also not provided, we try to add a simple dge

        if R1 == ["None"]:
            R1 = None

        if R2 == ["None"]:
            R2 = None

        if reads == "None":
            reads = None

        if action == "add" and (R2 is None) and (reads is None) and not is_merged:
            self.logger.info("R2 not provided, trying longreads")

            if not longreads:
                self.logger.info(
                    "longreads not provided, trying basecalls_dir and sample_sheet"
                )
                if basecalls_dir is None or sample_sheet is None:
                    self.logger.info(
                        "basecalls_dir or sample_sheet not provided, trying dge"
                    )
                    if not dge:
                        raise SpacemakeError(
                            "Neither R1,R2, longreads, basecalls_dir & "
                            + "sample_sheet, nor dge were provided.\n"
                            + "Some reads/data has to be provided"
                        )
            else:
                if not longread_signature:
                    raise SpacemakeError(
                        "adding longreads requires to set --longread-signature as well (e.g. dropseq, chromium, noUMI, default, visium, slideseq_bc14,...)"
                    )

        # assert files first
        # if R1 is not None:
        assert_file(R1, default_value=None, extension=".fastq.gz")

        # if R2 is not None:
        assert_file(R2, default_value=None, extension=".fastq.gz")

        assert_file(reads, default_value=None, extension=[".txt", ".csv", ".tsv"])
        assert_file(longreads, default_value=None, extension="all")
        assert_file(
            dge,
            default_value=None,
            extension=[".h5", ".csv", ".h5ad", ".loom", ".txt", ".txt.gz"],
        )

        # assign reads
        # if there are strings, make them lists, as one sample can have many read files
        if R1 is not None and isinstance(R1, str):
            R1 = [R1]

        if R2 is not None and isinstance(R2, str):
            R2 = [R2]

        if R1 is not None and R2 is not None:
            if len(R1) != len(R2):
                raise SpacemakeError(
                    f"Trying to set an unmatching number of "
                    + f"read pairs for sample: {ix}.\n"
                    + f"# of R1 files = {len(R1)}\n"
                    + f"# of R2 files = {len(R2)}"
                )

        if "run_mode" in kwargs and isinstance(kwargs["run_mode"], str):
            # if a single run mode is provided as a string
            # create a list manually
            kwargs["run_mode"] = [kwargs["run_mode"]]

        # check if run mode exists
        for run_mode in kwargs.get("run_mode", []):
            if not self.config.variable_exists("run_modes", run_mode):
                raise ConfigVariableNotFoundError("run_modes", run_mode)

        config_variables_to_check = {
            "pucks": "puck",
            "barcode_flavors": "barcode_flavor",
            "adapter_flavors": "adapter_flavor",
            "species": "species",
        }

        for cv_plural, cv_singular in config_variables_to_check.items():
            if cv_singular in kwargs.keys():
                if not self.config.variable_exists(cv_plural, kwargs[cv_singular]):
                    raise ConfigVariableNotFoundError(cv_singular, kwargs[cv_singular])

        # if everything correct, add or update
        # first populate kwargs
        kwargs["R1"] = R1
        kwargs["R2"] = R2
        kwargs["reads"] = reads
        kwargs["dge"] = dge
        kwargs["longreads"] = longreads
        kwargs["longread_signature"] = longread_signature
        kwargs["sample_sheet"] = sample_sheet
        kwargs["basecalls_dir"] = basecalls_dir
        kwargs["is_merged"] = is_merged

        if action == "add" and map_strategy is None:
            # was not specified! Let's evaluate the default for this species
            # (depends on having rRNA reference or not)
            map_strategy = self.get_default_map_strategy_for_species(kwargs["species"])

        if (map_strategy is not None) and (map_strategy.endswith("-")):
            self.logger.warning(
                (
                    f"\n!!!!WARNING!!!\n map_strategy '{map_strategy}' ends with a trailing dash."
                    " Please ensure map_strategy is escaped with double-quotes, or the shell"
                    " interpretes the chevron in '->' as a redirect."
                )
            )

        if action == "add":
            _i_species = kwargs["species"]
        elif action == "update":
            _i_species = self.df.loc[ix]["species"]

        # validate and correct map_strategies
        from spacemake.map_strategy import validate_mapstr

        kwargs["map_strategy"] = validate_mapstr(
            map_strategy, config=self.config, species=_i_species
        )

        # TODO: remove
        # kwargs["map_strategy"] = map_strategy

        # populate puck_barcode_file
        if puck_barcode_file is not None:
            if isinstance(puck_barcode_file, str):
                puck_barcode_file = [puck_barcode_file]

            # if there are duplicates, raise error
            if len(puck_barcode_file) != len(set(puck_barcode_file)):
                raise SpacemakeError(
                    "Duplicate files provided for "
                    + "--puck_barcode_file. \n"
                    + f"files provided: {puck_barcode_file}"
                )

        if puck_barcode_file_id is not None:
            if isinstance(puck_barcode_file_id, str):
                puck_barcode_file_id = [puck_barcode_file_id]

            if len(puck_barcode_file_id) != len(set(puck_barcode_file_id)):
                raise SpacemakeError(
                    "Duplicate ids provided for "
                    + "--puck_barcode_file_id. \n"
                    + f"ids provided: {puck_barcode_file_id}"
                )

        # if puck barcode files and id's provided
        if puck_barcode_file_id is not None and puck_barcode_file is not None:
            # checklist
            # check if the lengths are the same
            if len(puck_barcode_file_id) != len(puck_barcode_file):
                raise SpacemakeError(
                    "Unmatching number of arguments provided"
                    + " for --puck_barcode_file and --puck_barcode_file_id.\n"
                    + "The provided number of elements should be the same"
                )

            kwargs["puck_barcode_file_id"] = puck_barcode_file_id
            kwargs["puck_barcode_file"] = puck_barcode_file
        elif puck_barcode_file_id is None and puck_barcode_file is not None:
            # if the user only provided puck_barcode_files
            self.logger.info(
                "No --puck_barcode_file_id provided. Generating" + " ids from filename"
            )

            puck_barcode_file_id = [
                os.path.basename(path).split(".")[0] for path in puck_barcode_file
            ]

            # if duplicates detected here: different files but same basename
            # this can happen if the file extensions are different
            if len(puck_barcode_file_id) != len(set(puck_barcode_file_id)):
                for i in range(len(puck_barcode_file_id)):
                    puck_barcode_file_id[i] = f"{puck_barcode_file_id[i]}_{i}"

            kwargs["puck_barcode_file_id"] = puck_barcode_file_id
            kwargs["puck_barcode_file"] = puck_barcode_file

        else:
            # if no puck barcode files are provided, we check if the puck has barcodes
            puck_name = kwargs.get("puck", None)

            if puck_name is not None:
                puck = self.config.get_puck(puck_name)

                if puck.has_barcodes:
                    kwargs["puck_barcode_file_id"] = [puck_name]
                    kwargs["puck_barcode_file"] = puck.variables["barcodes"]

        if sample_exists:
            new_project = self.df.loc[ix].copy()
            # pd.Series.update will only update values which are not None
            new_project.update(pd.Series(kwargs))
            self.df.loc[ix] = new_project
        else:
            new_project = pd.Series(self.project_df_default_values)
            new_project.name = ix
            new_project.update(kwargs)

            # after addition
            self.df = pd.concat([self.df, pd.DataFrame(new_project).T], axis=0)
            # as of pandas 2.0.1 the names of a MultiIndex do not survive concat.
            self.df.index.names = ["project_id", "sample_id"]

        if return_series:
            return (ix, new_project)

    def delete_sample(self, project_id, sample_id):
        """delete_sample.

        :param project_id:
        :param sample_id:
        """
        ix = (project_id, sample_id)

        if self.sample_exists(*ix):
            element = self.df.loc[ix]

            self.logger.info(
                f"Deleting sample: {ix}, with the following"
                + f" with the following variables:\n{element}"
            )

            self.df.drop(ix, inplace=True)
            return element
        else:
            raise ProjectSampleNotFoundError(
                "(project_id, sample_id)", (project_id, sample_id)
            )

    def add_samples_from_yaml(self, projects_yaml_file):
        """add_samples_from_yaml.

        :param projects_yaml_file:
        """
        import yaml

        config = yaml.load(open(projects_yaml_file), Loader=yaml.FullLoader)
        demux_projects = config.get("projects", None)

        if demux_projects is not None:
            # if we have projects in the config file
            # get the samples
            for ip in demux_projects:
                self.add_sample_sheet(ip["sample_sheet"], ip["basecalls_dir"])

        # add additional samples from config.yaml, which have already been demultiplexed.
        for project in config["additional_projects"]:
            self.add_update_sample(**project)

    def get_ix_from_project_sample_list(self, project_id_list=[], sample_id_list=[]):
        """get_ix_from_project_sample_list.

        :param project_id_list:
        :param sample_id_list:
        """

        # raise error if both lists are empty
        if project_id_list == [] and sample_id_list == []:
            raise NoProjectSampleProvidedError()

        # of only one provided use that, if both use intersection
        if project_id_list == []:
            ix = self.df.query("sample_id in @sample_id_list").index
        elif sample_id_list == []:
            ix = self.df.query("project_id in @project_id_list").index
        else:
            ix = self.df.query(
                "project_id in @project_id_list and sample_id in @sample_id_list"
            ).index

        return ix

    def set_variable(self, ix, variable_name, variable_key, keep_old=False):
        """set_variable.

        :param ix:
        :param variable_name:
        :param variable_key:
        :param keep_old:
        """
        variable_name_pl = self.config.main_variables_sg2pl[variable_name]
        self.config.assert_main_variable(variable_name_pl)
        self.config.assert_variable(variable_name_pl, variable_key)

        # get current value
        i_variable = self.df.at[ix, variable_name]

        # if list, we either append or not
        if isinstance(i_variable, list):
            if keep_old:
                # keep the old list as well
                i_variable = list(set(i_variable + variable_key))
            else:
                i_variable = list(set(variable_key))

        # if we do not keep the old, simply set
        else:
            i_variable = variable_key

        if i_variable == [] or i_variable is None:
            raise EmptyConfigVariableError(variable_name)

        self.df.at[ix, variable_name] = i_variable

    def remove_variable(self, ix, variable_name, variable_key):
        """remove_variable.

        :param ix:
        :param variable_name:
        :param variable_key:
        """
        variable_name_pl = self.config.main_variables_sg2pl[variable_name]
        self.config.assert_main_variable(variable_name_pl)
        self.config.assert_variable(variable_name_pl, variable_key)

        if not isinstance(variable_key, list):
            raise ValueError("variable_key has to be a list")

        i_variable = self.df.at[ix, variable_name]

        i_variable = [val for val in i_variable if val not in variable_key]
        if i_variable == [] or i_variable is None:
            raise EmptyConfigVariableError(variable_name)

        self.df.at[ix, variable_name] = i_variable

    def set_remove_variable(
        self,
        variable_name,
        variable_key,
        action,
        project_id_list=[],
        sample_id_list=[],
        keep_old=False,
    ):
        """set_remove_variable.

        :param variable_name:
        :param variable_key:
        :param action:
        :param project_id_list:
        :param sample_id_list:
        :param keep_old:
        """
        self.assert_projects_samples_exist(project_id_list, sample_id_list)

        variable_name_pl = self.config.main_variables_sg2pl[variable_name]
        self.config.assert_main_variable(variable_name_pl)

        ix = self.get_ix_from_project_sample_list(
            project_id_list=project_id_list, sample_id_list=sample_id_list
        )

        if isinstance(variable_key, list):
            for var in variable_key:
                self.config.assert_variable(variable_name_pl, var)
        else:
            self.config.assert_variable(variable_name_pl, variable_key)

        for i, row in self.df.loc[ix, :].iterrows():
            # add/remove variable. if it already exists dont do anything
            if action == "set":
                self.set_variable(i, variable_name, variable_key, keep_old)
            elif action == "remove":
                self.remove_variable(i, variable_name, variable_key)

        return ix.to_list()

    def assert_projects_samples_exist(self, project_id_list=[], sample_id_list=[]):
        """assert_projects_samples_exist.

        :param project_id_list:
        :param sample_id_list:
        """
        for project in project_id_list:
            if project not in self.df.index.get_level_values("project_id"):
                raise ProjectSampleNotFoundError("project_id", project)

        for sample in sample_id_list:
            if sample not in self.df.index.get_level_values("sample_id"):
                raise ProjectSampleNotFoundError("sample_id", sample)

    def merge_samples(
        self,
        merged_project_id,
        merged_sample_id,
        project_id_list=[],
        sample_id_list=[],
        **kwargs,
    ):
        """merge samples.

        :param merged_project_id:
        :param merged_sample_id:
        :param project_id_list:
        :param sample_id_list:
        :param kwargs:
        """
        # check if projects and samples with these IDs exist
        self.assert_projects_samples_exist(project_id_list, sample_id_list)

        ix = self.get_ix_from_project_sample_list(
            project_id_list=project_id_list, sample_id_list=sample_id_list
        )

        consistent_variables = list(self.config.main_variables_sg2pl.keys())
        consistent_variables.remove("run_mode")
        consistent_variables.remove("adapter")
        consistent_variables.remove("quant")

        # append variable "map-strategy" manually.
        # TODO: consider moving "map_strategy" into the main_variables_sg2pl, needs additional parser etc.
        consistent_variables.append("map_strategy")

        ix_list = ix.to_list()

        if ix_list == []:
            raise ProjectSampleNotFoundError(
                "(project_id_list, sample_id_list)", (project_id_list, sample_id_list)
            )

        # check for variable inconsistency
        # raise error if variable different between samples
        for variable in consistent_variables:
            if variable in kwargs and variable not in ["species", "barcode_flavor"]:
                # if variable provided from command line, skip
                self.logger.info(f"{variable} provided, skipping deduction...")
                continue

            self.logger.info(
                f"{variable} not provided. deducing from merged samples..."
            )

            variable_val = self.df.loc[ix, variable].to_list()

            if len(set(variable_val)) > 1:
                raise InconsistentVariablesDuringMerge(
                    variable_name=variable, variable_value=variable_val, ix=ix
                )
            else:
                # attach the deduced, consisten variable
                kwargs[variable] = variable_val[0]

        # get puck_barcode_files
        if "puck_barcode_file" not in kwargs or "puck_barcode_file_id" not in kwargs:
            kwargs["puck_barcode_file_id"] = []
            kwargs["puck_barcode_file"] = []

            for _, row in self.df.loc[ix].iterrows():
                if row["puck_barcode_file"] is None:
                    continue
                else:
                    for pbf_id, pbf in zip(
                        row["puck_barcode_file_id"], row["puck_barcode_file"]
                    ):
                        if (
                            pbf_id not in kwargs["puck_barcode_file_id"]
                            and pbf not in kwargs["puck_barcode_file"]
                        ):
                            kwargs["puck_barcode_file_id"].append(pbf_id)
                            kwargs["puck_barcode_file"].append(pbf)

        # after all checks, log that we are merging
        self.logger.info(f"Merging samples {ix_list} together\n")
        variables_to_deduce = ["investigator", "experiment", "sequencing_date"]

        for variable in variables_to_deduce:
            if variable not in kwargs:
                kwargs[variable] = ";".join(self.df.loc[ix, variable].unique())

        # if no run_mode provided, overwrite with user defined one
        if "run_mode" not in kwargs.keys():
            run_mode_lists = [
                set(run_mode) for run_mode in self.df.loc[ix].run_mode.to_list()
            ]

            # join run modes from parent samples
            if len(run_mode_lists) == 1:
                run_mode = run_mode_lists[0]
            else:
                run_mode = run_mode_lists[0].intersection(*run_mode_lists[1:])

            # create a list from the set intersection
            run_mode = list(run_mode)

            # if there are no common elements, throw an error
            if len(run_mode) == 0:
                raise InconsistentVariablesDuringMerge(
                    variable_name="run_mode", variable_value=run_mode_lists, ix=ix_list
                )

            # finally add run mode to arguments
            kwargs["run_mode"] = run_mode

        # set the action to add
        kwargs["action"] = "add"

        sample_added = self.add_update_sample(
            project_id=merged_project_id,
            sample_id=merged_sample_id,
            is_merged=True,
            merged_from=ix.to_list(),
            **kwargs,
        )

        return (sample_added, ix)


__global_ProjectDF = None


def get_global_ProjectDF(root="."):
    global __global_ProjectDF
    if __global_ProjectDF is None:
        from spacemake.config import get_global_config

        __global_ProjectDF = ProjectDF(
            f"{root}/project_df.csv", config=get_global_config()
        )

    return __global_ProjectDF



__version__ = "0.9"
__author__ = ["Marvin Jens"]
__license__ = "GPL"
__email__ = ["marvin.jens@mdc-berlin.de"]

import logging
import time


def put_or_abort(Q, item, abort_flag, timeout=1):
    """
    Small wrapper around queue.put() to prevent
    dead-locks in the event of (detectable) errors
    that might cause put() to block forever.
    Expects a shared mp.Value instance as abort_flag

    Returns: False if put() was succesful, True if execution
    should be aborted.
    """
    import queue

    sent = False
    # logging.warning(f"sent={sent} abort_flag={abort_flag}")
    while not (sent or abort_flag.value):
        try:
            Q.put(item, timeout=timeout)
        except queue.Full:
            pass
        else:
            sent = True

    return abort_flag.value


def queue_iter(Q, abort_flag, stop_item=None, timeout=1):
    """
    Small generator/wrapper around multiprocessing.Queue allowing simple
    for-loop semantics:

        for item in queue_iter(queue, abort_flag):
            ...
    The abort_flag is handled analogous to put_or_abort, only
    that it ends the iteration instead
    """
    import queue

    # logging.debug(f"queue_iter({queue})")
    while True:
        if abort_flag.value:
            break
        try:
            item = Q.get(timeout=timeout)
        except queue.Empty:
            pass
        else:
            if item == stop_item:
                # signals end->exit
                break
            else:
                # logging.debug(f"queue_iter->item {item}")
                yield item


def join_with_empty_queues(proc, Qs, abort_flag, timeout=1):
    """
    joins() a process that writes data to queues Qs w/o deadlock.
    In case of an abort, the subprocess normally would not join
    until the Qs are emptied. join_with_empty_queues() monitors a global
    abort flag and empties the queues if needed, allowing the sub-process
    to terminate properly.
    """

    def drain(Q):
        content = []
        while not Q.empty():
            try:
                item = Q.get(timeout=timeout)
            except queue.Empty:
                pass
            else:
                content.append(item)

        return content

    contents = [list() for i in range(len(Qs))]
    while proc.exitcode is None:
        proc.join(timeout)
        if abort_flag.value:
            for Q, content in zip(Qs, contents):
                content.extend(drain(Q))

    return contents


def chunkify(src, n_chunk=1000):
    """
    Iterator which collects up to n_chunk items from iterable <src> and yields them
    as a list.
    """
    chunk = []
    n = 0
    for x in src:
        chunk.append(x)
        if len(chunk) >= n_chunk:
            yield n, chunk
            n += 1
            chunk = []

    if chunk:
        yield n, chunk


def log_qerr(qerr):
    "helper function for reporting errors in sub processes"
    for name, lines in qerr:
        for line in lines:
            logging.error(f"subprocess {name} exception {line}")


class ExceptionLogging:
    """
    A context manager that handles otherwise uncaught exceptions by logging
    the event and traceback info, optinally raises a flag.
    Very handy for wrapping the main function in a sub-process!
    """

    def __init__(self, name, Qerr=None, exc_flag=None):
        # print('__init__ called')
        self.Qerr = Qerr
        self.exc_flag = exc_flag
        self.name = name
        self.logger = logging.getLogger(name)
        self.exception = None

    def __enter__(self):
        self.t0 = time.time()
        # print('__enter__ called')
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        # print('__exit__ called')
        self.t1 = time.time()
        self.logger.info(f"CPU time: {self.t1 - self.t0:.3f} seconds.")
        if exc_type and (exc_type != SystemExit):
            import traceback

            lines = "\n".join(
                traceback.format_exception(exc_type, exc_value, exc_traceback)
            ).split("\n")
            self.exception = lines
            self.logger.error(f"an unhandled exception occurred")
            for l in lines:
                self.logger.error(l)

            if self.Qerr is not None:
                self.Qerr.put((self.name, lines))

            if self.exc_flag:
                self.logger.error(f"raising exception flag {self.exc_flag}")
                self.exc_flag.value = True



import logging

import multiprocessing as mp
import pandas as pd
import time

from spacemake.util import message_aggregation, FASTQ_src

"""
Sequence intersection between a query file and target files (pucks).

Usage:
    python create_spatial_barcode_file.py \
        --query input_reads.txt \
        --target puck_file1.txt puck_file2.txt \
        --output puck_output1.txt puck_output2.txt

Author:
    Daniel León-Periñán
"""

logger_name = "spacemake.snakemake.scripts.create_spatial_barcode_file"
logger = logging.getLogger(logger_name)
AVAILABLE_CPUS = mp.cpu_count()


def setup_parser(parser):
    """
    Set up command-line arguments for the script.

    :param parser: Argument parser object.
    :type parser: argparse.ArgumentParser
    :returns: Updated argument parser object.
    :rtype: argparse.ArgumentParser
    """
    parser.add_argument(
        "--query",
        type=str,
        help="query reads file for lookup into 'target'(s) database",
        default=None,
    )

    parser.add_argument(
        "--query-tag",
        type=str,
        help="if input is a bam file, the specified tag will be used as the sequence",
        default="CB",
    )

    parser.add_argument(
        "--query-plain-skip",
        type=int,
        help="number of lines to skip, if target input is a plain text file",
        default=None,
    )

    parser.add_argument(
        "--query-plain-column",
        type=int,
        help="column to parse from query, if input is a plain text file",
        default=None,
    )

    parser.add_argument(
        "--query-separator",
        type=str,
        default="\t",
        help="if input is a tabular plain text file, this specifies the character for column separation",
        choices=["\t", ",", ";"],
    )

    parser.add_argument(
        "--target",
        type=str,
        nargs="+",
        help="target puck files against which search is performed",
        required=True,
    )

    parser.add_argument(
        "--target-id",
        type=str,
        nargs="+",
        help="target puck ids for summary output",
        required=True,
    )

    parser.add_argument(
        "--target-tag",
        type=str,
        help="if input is a bam file, the specified tag will be used as the sequence",
        default="CB",
    )

    parser.add_argument(
        "--target-plain-skip",
        type=int,
        help="number of lines to skip, if target input is a plain text file",
        default=None,
    )

    parser.add_argument(
        "--target-plain-column",
        type=int,
        help="column to parse from target, if input is a plain text file",
        default=None,
    )

    parser.add_argument(
        "--target-separator",
        type=str,
        default="\t",
        help="if input is a tabular plain text file, this specifies the character for column separation",
        choices=["\t", ",", ";"],
    )

    parser.add_argument(
        "--summary-output",
        type=str,
        help="a summary output file containing the number of matches between query and target, per target file",
        required=True,
    )

    parser.add_argument(
        "--min-threshold",
        type=float,
        default=0,
        help="targets with 'matching_ratio' above --min-threshold are marked as True; else, False.",
        required=False,
    )

    parser.add_argument(
        "--n-jobs",
        type=int,
        help="number of parallel jobs (up to {AVAILABLE_CPUS} in this machine) for database lookup",
        metavar=f"[0-{AVAILABLE_CPUS-1}]",
        choices=range(0, AVAILABLE_CPUS),
        default=1,
    )

    return parser


def BAM_src(src: str, tag: str = None) -> tuple:
    """
    Generate read data from a BAM file.

    :param src: Path to the BAM file.
    :type src: str
    :param tag: Tag to be used as the sequence. Defaults to None.
    :type tag: str, optional
    :yields: Tuple containing read name, sequence, and quality.
    :rtype: tuple
    """

    import pysam

    bam = pysam.AlignmentFile(src, "rb", check_sq=False)
    for read in bam.fetch(until_eof=True):
        if tag is None:
            yield read.query_name, read.query_sequence, read.query_qualities
        else:
            yield None, read.get_tag(tag), None


def plain_src(f: str, skip: int = 0, column: int = 0, separator: str = "\t") -> tuple:
    """
    Generate read data from a plain text file.

    :param f: Path to the plain text file.
    :type f: str
    :param skip: Number of lines to skip. Defaults to 0.
    :type skip: int, optional
    :param column: Column to parse from the file. Defaults to 0.
    :type column: int, optional
    :param separator: Character for column separation. Defaults to "\t".
    :type separator: str, optional
    :yields: Tuple containing read name, sequence, and quality.
    :rtype: tuple
    """
    
    reads = pd.read_csv(f, sep=separator, skiprows=skip)
    reads = reads.iloc[:, column]

    for read in reads:
        yield None, read, None


def read_with_tag(
    f: str, tag: str = None, skip: int = None, column: int = 0, separator: str = "\t"
) -> str:  
    """
    Read data from various sources based on the file type.

    :param f: Path to the input file.
    :type f: str
    :param tag: Tag to be used as the sequence. Defaults to None.
    :type tag: str, optional
    :param skip: Number of lines to skip (plain text files only). Defaults to None.
    :type skip: int, optional
    :param column: Column to parse from the file (plain text files only). Defaults to 0.
    :type column: int, optional
    :param separator: Character for column separation (plain text files only). Defaults to "\t".
    :type separator: str, optional
    :yields: Sequence from the input file.
    :rtype: str
    """
    import gzip

    if type(f) is not str:
        src = FASTQ_src(f)  # assume its a stream or file-like object already
    elif f.endswith(".txt") or f.endswith(".txt.gz"):
        src = plain_src(f, skip, column, separator)
    elif f.endswith(".gz"):
        src = FASTQ_src(gzip.open(f, mode="rt"))
    elif f.endswith(".bam"):
        src = BAM_src(f, tag=tag)
    else:
        src = FASTQ_src(open(f))

    for _, seq, _ in src:
        yield seq


def read_plain_src_chunked(
    f: str, tag: str = None, skip: int = None, column: int = 0, separator: str = "\t", chunk_size=100_000_000
):  
    """
    Read data from various sources based on the file type.

    :param f: Path to the input file.
    :type f: str
    :param tag: Tag to be used as the sequence. Defaults to None.
    :type tag: str, optional
    :param skip: Number of lines to skip (plain text files only). Defaults to None.
    :type skip: int, optional
    :param column: Column to parse from the file (plain text files only). Defaults to 0.
    :type column: int, optional
    :param separator: Character for column separation (plain text files only). Defaults to "\t".
    :type separator: str, optional
    :yields: Sequence from the input file.
    :rtype: str
    """

    chunks = pd.read_csv(f, sep=separator, skiprows=skip, chunksize=chunk_size)

    for chunk in chunks:
        reads = chunk.iloc[:, column]
        yield reads


# TODO: use a partial function to pass args to not use global context
def find_matches(f: str):
    """
    Find matches between query and target sequences.

    :param f: Target file to compare against query sequences.
    :type f: str
    :returns: Tuple containing information about the target file, number of barcodes, number of matching barcodes, and the matching ratio.
    :rtype: tuple
    """
    global query_seqs, args

    start = time.time()
    target = set(
        read_with_tag(
            f,
            args.target_tag,
            args.target_plain_skip,
            args.target_plain_column,
            args.target_separator,
        )
    )
    n_matches = len(target.intersection(query_seqs))
    pct_matches_target = n_matches / len(target)
    print(f"queried {f:,} in {round(time.time()-start, 2):,} seconds")

    return (f, len(target), n_matches, pct_matches_target)

# TODO: implement chunked so it only reads 100M reads every time
@message_aggregation(logger_name)
def cmdline():
    """cmdline."""
    import argparse

    global args, query_seqs

    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description="calculate sequence intersection between a query and target(s) file(s)",
    )
    parser = setup_parser(parser)

    args = parser.parse_args()

    if len(args.target_id) != len(args.target):
        raise ValueError(
            f"target_id ({len(args.target_id)}) and target ({len(args.target)}) are different in size"
        )

    query_seqs = read_with_tag(
        args.query,
        args.query_tag,
        args.query_plain_skip,
        args.query_plain_column,
        args.query_separator,
    )
    logger.info(f"read query file from '{args.query}'")

    # get unique reads into a global context
    query_seqs = set(query_seqs)
    logger.info(f"hashed unique query reads")
    logger.info(
        f"querying against {len(args.target)} targets with {args.n_jobs} parallel jobs"
    )

    with mp.Pool(args.n_jobs) as pool:
        results = pool.map(find_matches, args.target)

    df = pd.DataFrame(
        results,
        columns=["puck_barcode_file", "n_barcodes", "n_matching", "matching_ratio"],
    )
    df["puck_barcode_file_id"] = args.target_id
    df['pass_threshold'] = 0
    df['pass_threshold'][df['matching_ratio'] > args.min_threshold] = 1
    df.to_csv(args.summary_output, index=False)


if __name__ == "__main__":
    cmdline()



import os

# import yaml
import argparse
from spacemake.contrib import __version__
import re
import logging

from spacemake.errors import *
from spacemake.util import str2bool, assert_file, bool_in_str, message_aggregation
from spacemake.util import check_star_index_compatibility

logger_name = "spacemake.config"


__global_config = None


def get_global_config(root="."):
    global __global_config
    if __global_config is None:
        __global_config = ConfigFile.from_yaml(f"{root}/config.yaml")

    return __global_config


def get_puck_parser(required=True):
    parser = argparse.ArgumentParser(allow_abbrev=False, add_help=False)
    parser.add_argument("--name", help="name of the puck", type=str, required=True)
    parser.add_argument(
        "--width_um", type=float, required=required, help="width of the puck in microns"
    )
    parser.add_argument(
        "--spot_diameter_um",
        type=float,
        required=required,
        help="diameter of the spots in this puck, in microns",
    )
    parser.add_argument(
        "--barcodes",
        type=str,
        required=False,
        help="path to barcode file. if not provided, the --puck_barcode_file variable"
        + " of `spacemake projects add_sample` has to be set",
    )
    parser.add_argument(
        "--coordinate_system",
        type=str,
        required=False,
        help="path to coordinate system file. When specified, spacemake will 'stitch'"
        + " pucks into a single file, with corresponding global coordinates",
    )

    return parser


def get_run_mode_parser(required=True):
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        formatter_class=argparse.RawTextHelpFormatter,
        description="add/update run_mode parent parser",
        add_help=False,
    )
    parser.add_argument(
        "--name", type=str, help="name of the run_mode to be added", required=True
    )
    parser.add_argument(
        "--parent_run_mode",
        type=str,
        help="Name of the parent run_mode. All run_modes will fall back to 'default'",
    )
    parser.add_argument(
        "--umi_cutoff",
        type=int,
        nargs="+",
        help="umi_cutoff for this run_mode."
        + "the automated analysis will be run with these cutoffs",
    )
    parser.add_argument(
        "--n_beads", type=int, help="number of expected beads for this run_mode"
    )
    parser.add_argument(
        "--clean_dge",
        required=False,
        choices=bool_in_str,
        type=str,
        help="if True, the DGE will be cleaned of barcodes which overlap with primers",
    )
    parser.add_argument(
        "--detect_tissue",
        required=False,
        choices=bool_in_str,
        type=str,
        help="By default only beads having at least umi_cutoff UMI counts are analysed "
        + "during the automated analysis, all other beads are filtered out. If this "
        + "parameter is set, contiguous islands within umi_cutoff passing beads will "
        + "also be included in the analysis",
    )
    # parser.add_argument(
    #     "--polyA_adapter_trimming",
    #     required=False,
    #     choices=bool_in_str,
    #     type=str,
    #     help="if set, reads will have polyA stretches and adapter sequence overlaps trimmed "
    #     + "BEFORE mapping.",
    # )
    parser.add_argument(
        "--count_intronic_reads",
        required=False,
        choices=bool_in_str,
        type=str,
        help="if set, INTRONIC reads will also be countsed (apart from UTR and CDS)",
    )
    parser.add_argument(
        "--count_mm_reads",
        required=False,
        choices=bool_in_str,
        type=str,
        help="if True, multi-mappers will also be counted. For every "
        + "multimapper only reads which have one unique read mapped"
        + "to a CDS or UTR region will be counted",
    )

    parser.add_argument(
        "--mesh_data",
        required=False,
        choices=bool_in_str,
        type=str,
        help="if True, this data will be 'mehsed': a hexagonal structured "
        + "meshgrid will be created, where each new spot will have diameter"
        + " of --mesh_spot_diameter_um micron diameter and the spots will "
        + "be spaced --mesh_spot_distance_um microns apart",
    )
    parser.add_argument(
        "--mesh_type",
        required=False,
        choices=["circle", "hexagon"],
        type=str,
        help="circle: circles with diameter of --mesh_spot_diameter_um will be placed"
        + " in a hex grid, where he distance between any two circles will be "
        + "--mesh_spot_distance_um\n"
        + "hexagon: a mesh of touching hexagons will be created, with their "
        + "centers being --mesh_spot_distance_um apart. This will cover all "
        + "the data without any holes",
    )
    parser.add_argument(
        "--mesh_spot_diameter_um",
        type=float,
        required=False,
        help="diameter of mesh spot, in microns. to create a visium-style "
        + "mesh, use 55um",
    )
    parser.add_argument(
        "--mesh_spot_distance_um",
        type=float,
        required=False,
        help="distance between mesh spots in um. to create a visium-style "
        + "mesh use 100um",
    )
    parser.add_argument(
        "--spatial_barcode_min_matches",
        type=float,
        required=False,
        help="minimum ratio (0, 1] of spatial barcode matches to further consider a puck"
        + "across the rest of the pipeline",
    )

    return parser


def get_species_parser(required=True):
    "a parser that allows to add a reference sequence and annotation, belonging to some species"
    parser = argparse.ArgumentParser(allow_abbrev=False, add_help=False)
    parser.add_argument(
        "--reference",
        help="name of the reference (default=genome)",
        type=str,
        default="genome",
    )
    parser.add_argument("--name", help="name of the species", type=str, required=True)
    parser.add_argument(
        "--sequence",
        help="path to the sequence (.fa) file for the species/reference to be added (e.g. the genome)",
        type=str,
        required=required,
    )
    parser.add_argument(
        "--genome",
        help="[DEPRECATED] path to the genome (.fa) file for the species to be added. --genome=<arg> is a synonym for --reference=genome --sequence=<arg>",
        type=str,
        required=False,
    )

    parser.add_argument(
        "--annotation",
        help="path to the genome annotation (.gtf) file for the species to be added",
        type=str,
        default="",
        required=False,
    )
    parser.add_argument(
        "--STAR_index_dir",
        help="path to STAR index directory",
        type=str,
        required=False,
    )
    parser.add_argument(
        "--BT2_index",
        help="path to BOWTIE2 index",
        type=str,
        required=False,
    )
    parser.add_argument(
        "--BT2_flags",
        help="bt2 mapping arguments for this reference (default=mapping.smk:default_BT2_MAP_FLAGS) ",
        type=str,
        default="",
        required=False,
    )
    parser.add_argument(
        "--STAR_flags",
        help="STAR mapping arguments for this reference (default=mapping.smk:default_STAR_MAP_FLAGS)",
        type=str,
        default="",
        required=False,
    )

    return parser


def get_barcode_flavor_parser(required=True):
    parser = argparse.ArgumentParser(
        allow_abbrev=False, description="add/update barcode-flavor", add_help=False
    )
    parser.add_argument(
        "--name", help="name of the barcode flavor", type=str, required=True
    )
    parser.add_argument(
        "--umi",
        help="structure of UMI, using python's list syntax. Example: to set UMI to "
        + "13-20 NT of Read1, use --umi r1[12:20]. It is also possible to use the first 8nt of "
        + "Read2 as UMI: --umi r2[0:8]",
        type=str,
        required=required,
    )
    parser.add_argument(
        "--cell-barcode",
        help="structure of CELL BARCODE, using python's list syntax. Example: to set"
        + " the cell_barcode to 1-12 nt of Read1, use --cell-barcode r1[0:12]. It is also possible "
        + " to reverse the CELL BARCODE, for instance with r1[0:12][::-1] (reversing the first 12nt of"
        + " Read1, and assigning them as CELL BARCODE).",
        type=str,
        required=required,
    )
    parser.add_argument(
        "--cell_barcode",
        help=argparse.SUPPRESS,
        type=str,
    )

    return parser


def get_variable_action_subparsers(config, parent_parser, variable):
    if variable == "barcode_flavors":
        variable_singular = variable[:-1]
        variable_add_update_parser = get_barcode_flavor_parser
    elif variable == "pucks":
        variable_singular = variable[:-1]
        variable_add_update_parser = get_puck_parser
    elif variable == "run_modes":
        variable_singular = variable[:-1]
        variable_add_update_parser = get_run_mode_parser
    elif variable == "species":
        variable_singular = variable
        variable_add_update_parser = get_species_parser

    command_help = {
        "list": f"list {variable} and their settings",
        "delete": f"delete {variable_singular}",
        "add": f"add a new {variable_singular}",
        "update": f"update an existing {variable_singular}",
    }

    # list command
    list_parser = parent_parser.add_parser(
        f"list_{variable}", description=command_help["list"], help=command_help["list"]
    )
    list_parser.set_defaults(
        func=lambda args: list_variables_cmdline(config, args), variable=variable
    )

    func = lambda args: add_update_delete_variable_cmdline(config, args)

    # delete command
    delete_parser = parent_parser.add_parser(
        f"delete_{variable_singular}",
        description=command_help["delete"],
        help=command_help["delete"],
    )
    delete_parser.add_argument(
        "--name",
        help=f"name of the {variable_singular} to be deleted",
        type=str,
        required=True,
    )
    if variable == "species":
        delete_parser.add_argument(
            "--reference",
            help=f"name of the reference to be deleted (genome, rRNA, ...)",
            type=str,
            required=True,
        )

    delete_parser.set_defaults(func=func, action="delete", variable=variable)

    # add command
    add_parser = parent_parser.add_parser(
        f"add_{variable_singular}",
        parents=[variable_add_update_parser()],
        description=command_help["add"],
        help=command_help["add"],
    )
    add_parser.set_defaults(func=func, action="add", variable=variable)

    # update command
    update_parser = parent_parser.add_parser(
        f"update_{variable_singular}",
        parents=[variable_add_update_parser(False)],
        description=command_help["update"],
        help=command_help["update"],
    )
    update_parser.set_defaults(func=func, action="update", variable=variable)


def setup_config_parser(config, parent_parser_subparsers):
    parser_config = parent_parser_subparsers.add_parser(
        "config", help="configure spacemake"
    )
    parser_config_subparsers = parser_config.add_subparsers(
        help="config sub-command help"
    )

    for variable in ConfigFile.main_variables_pl2sg.keys():
        get_variable_action_subparsers(config, parser_config_subparsers, variable)

    return parser_config


@message_aggregation(logger_name)
def add_update_delete_variable_cmdline(config, args):
    # set the name and delete from dictionary
    name = args["name"]
    variable = args["variable"]
    action = args["action"]

    config.assert_main_variable(variable)

    # remove the args from the dict
    del args["action"]
    del args["variable"]
    del args["name"]

    if action == "add":
        func = config.add_variable
    elif action == "update":
        func = config.update_variable
    elif action == "delete":
        func = config.delete_variable

    var_variables = func(variable, name, **args)
    # print and dump config file
    config.logger.info(yaml.dump(var_variables, sort_keys=False))
    config.dump()


@message_aggregation(logger_name)
def list_variables_cmdline(config, args):
    variable = args["variable"]
    del args["variable"]
    import yaml

    config.logger.info(f"Listing {variable}")
    config.logger.info(yaml.dump(config.variables[variable]))


class ConfigMainVariable:
    def __init__(self, name, **kwargs):
        self.name = name
        self.variables = {}

        # loop over kwargs
        for variable_key, variable in kwargs.items():
            if variable_key not in self.variable_types:
                raise UnrecognisedConfigVariable(
                    f"{variable_key}", list(self.variable_types.keys())
                )
            elif self.variable_types[variable_key] == "int_list":
                self.variables[variable_key] = [int(x) for x in kwargs[variable_key]]
            else:
                new_type = self.variable_types[variable_key]
                self.variables[variable_key] = new_type(kwargs[variable_key])

    def __str__(self):
        class_name = self.__class__.__name__
        return f"{class_name}: {self.name}. variables: {self.variables}"

    def update(self, other):
        self.variables.update(other.variables)


class RunMode(ConfigMainVariable):
    variable_types = {
        "parent_run_mode": str,
        "n_beads": int,
        "umi_cutoff": "int_list",
        "clean_dge": bool,
        "detect_tissue": bool,
        "polyA_adapter_trimming": bool,
        "count_mm_reads": bool,
        "count_intronic_reads": bool,
        "mesh_data": bool,
        "mesh_type": str,
        "mesh_spot_diameter_um": int,
        "mesh_spot_distance_um": int,
        "spatial_barcode_min_matches": float,
    }

    def has_parent(self):
        if "parent_run_mode" in self.variables.keys():
            return True
        else:
            return False

    @property
    def parent_name(self):
        if self.has_parent():
            return self.variables["parent_run_mode"]
        else:
            return None


class Puck(ConfigMainVariable):
    variable_types = {
        "barcodes": str,
        "spot_diameter_um": float,
        "width_um": int,
        "coordinate_system": str,
    }

    @property
    def has_barcodes(self):
        return (
            "barcodes" in self.variables
            and self.variables["barcodes"]
            and self.variables["barcodes"] != "None"
        )

    @property
    def has_coordinate_system(self):
        return (
            "coordinate_system" in self.variables
            and self.variables["coordinate_system"]
            and self.variables["coordinate_system"] != "None"
        )


class ConfigFile:
    initial_config_path = os.path.join(
        os.path.dirname(__file__), "data/config/config.yaml"
    )

    main_variables_pl2sg = {
        "pucks": "puck",
        "barcode_flavors": "barcode_flavor",
        "adapters": "adapter",
        "adapter_flavors": "adapter_flavor",
        "run_modes": "run_mode",
        "species": "species",
        "quant": "quant",
    }

    main_variables_sg2pl = {value: key for key, value in main_variables_pl2sg.items()}

    main_variable_sg2type = {
        "puck": str,
        "barcode_flavor": str,
        "adapter": str,
        "adapter_flavor": str,
        "run_mode": "str_list",
        "species": str,
        "map_strategy": str,
    }

    def __init__(self):
        self.variables = {
            "spacemake_version": __version__,
            "root_dir": ".",
            "temp_dir": "/tmp",
            "species": {},
            "barcode_flavors": {},
            "adapter_flavors": {},
            "adapters": {},
            "run_modes": {},
            "pucks": {},
            "quant": {},
        }
        self.file_path = "config.yaml"
        self.logger = logging.getLogger(logger_name)

    @classmethod
    def from_yaml(cls, file_path="config.yaml"):
        cf = cls()
        import yaml

        config_yaml_variables = None
        with open(file_path, "r") as f:
            config_yaml_variables = yaml.load(f, Loader=yaml.FullLoader)

        if config_yaml_variables is not None:
            cf.variables.update(config_yaml_variables)

        cf.file_path = file_path

        if file_path != cf.initial_config_path:
            initial_config = ConfigFile.from_yaml(cf.initial_config_path)

            # correct variables to ensure backward compatibility
            cf.correct()
            # check which variables do not exist, if they dont,
            # copy them from initial config
            for main_variable in cf.main_variables_pl2sg:
                # update new main_variables
                if main_variable not in cf.variables:
                    cf.variables[main_variable] = initial_config.variables[
                        main_variable
                    ]

            # deduce variables which have 'default' value. this is to ensure spacemake
            # always runs w/o errors downstream: ie when barcode flavor, run_mode or puck
            # is set to default
            cf.vars_with_default = [
                key
                for key, value in initial_config.variables.items()
                if "default" in value
            ]

            for var_with_default in cf.vars_with_default:
                default_val = initial_config.variables[var_with_default]["default"]
                if "default" not in cf.variables[var_with_default]:
                    cf.variables[var_with_default]["default"] = default_val
                else:
                    user_val = cf.variables[var_with_default]["default"]
                    # update default run mode with missing values
                    if type(default_val) is dict:
                        default_val.update(user_val)
                    elif type(default_val) is list:
                        if type(user_val) is list:
                            default_val = user_val
                        else:
                            cf.logger.warning(
                                f"Trying to set list type entry {var_with_default}['default'] "
                                f"with a dict-type entry in user-defined '{file_path}'. "
                                "it looks like you may have a legacy config.yaml. Reverting to default values. "
                                "Please run 'spacemake migrate config' to get rid of this warning."
                            )
                    # print(default_val)
                    cf.variables[var_with_default]["default"] = default_val

        cf.expand_strings()
        return cf

    def expand_strings(self, **kw):
        kw["spacemake_dir"] = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "../")
        )
        # print(f"expanding strings in config with context={kw}")
        import collections.abc as abc

        def recurse(d):
            if type(d) == str:
                try:
                    return d.format(**kw)
                except KeyError:
                    return d

            if isinstance(d, abc.Mapping):
                for key, value in d.items():
                    d[key] = recurse(value)

            elif isinstance(d, abc.Collection):
                d = [recurse(x) for x in d]

            return d

        self.variables = recurse(self.variables)

    def assert_main_variable(self, variable):
        if variable not in self.main_variables_pl2sg.keys():
            raise UnrecognisedConfigVariable(
                variable, list(self.main_variables_pl2sg.keys())
            )

    def correct(self):
        # ensures backward compatibility
        if "pucks" not in self.variables:
            self.variables["pucks"] = {}
            if "puck_data" in self.variables:
                if "pucks" in self.variables["puck_data"]:
                    self.variables["pucks"] = self.variables["puck_data"]["pucks"]
                    del self.variables["puck_data"]["pucks"]

        if "barcode_flavors" not in self.variables:
            self.variables["barcode_flavors"] = {}
            if "knowledge" in self.variables:
                if "barcode_flavor" in self.variables["knowledge"]:
                    self.variables["barcode_flavors"] = self.variables["knowledge"][
                        "barcode_flavor"
                    ]

        if "species" not in self.variables:
            # get all the species and save them in the right place
            # if species is empty, create a species dictionary
            self.variables["species"] = {}

            # if "knowledge" in self.variables:
            #     # TODO: check if this still applies and make compatible with new
            #     # two-layer species->reference model
            #     # extract all annotation info, if exists
            #     for species in self.variables["knowledge"].get("annotations", {}):
            #         if species not in self.variables["species"]:
            #             self.variables["species"][species] = {}

            #         self.variables["species"][species]["annotation"] = self.variables[
            #             "knowledge"
            #         ]["annotations"][species]

            #     for species in self.variables["knowledge"].get("genomes", {}):
            #         if species not in self.variables["species"]:
            #             self.variables["species"][species] = {}

            #         self.variables["species"][species]["genome"] = self.variables[
            #             "knowledge"
            #         ]["genomes"][species]

            #     for species in self.variables["knowledge"].get("rRNA_genomes", {}):
            #         if species not in self.variables["species"]:
            #             self.variables["species"][species] = {}

            #         self.variables["species"][species]["rRNA_genome"] = self.variables[
            #             "knowledge"
            #         ]["rRNA_genomes"][species]

        for name, species_data in self.variables["species"].items():
            if "annotation" in species_data:
                # we have a pre-bowtie2-support config.yaml with following structur
                # species:
                #   human:
                #     genome: ....
                #     annotation: ....
                self.logger.warning(
                    f"converting old-style config.yaml species section for '{name}' {species_data}..."
                )
                new_data = {}
                new_data["genome"] = {
                    "sequence": species_data["genome"],
                    "annotation": species_data["annotation"],
                }
                if "rRNA_genome" in species_data:
                    new_data["rRNA"] = {
                        "sequence": species_data["rRNA_genome"],
                        "annotation": "",
                    }

                self.variables["species"][name] = new_data

        if "knowledge" in self.variables:
            del self.variables["knowledge"]

        # correct run modes
        for run_mode_name, run_mode_variables in self.variables["run_modes"].items():
            variables = run_mode_variables.copy()
            for var in run_mode_variables:
                if not var in RunMode.variable_types:
                    del variables[var]

            if ("polyA_adapter_trimming" in variables) and (
                variables["polyA_adapter_trimming"] == False
            ):
                import logging

                logger = logging.getLogger("spacemake.config")
                logger.warning(
                    f"WARNING: run_mode {run_mode_name} lists polyA_adapter_trimming=false. This is no longer supported and will be overriden with true"
                )
                variables["polyA_adapter_trimming"] = True

            # print(f"assigning run mode {run_mode_name}: {variables}")
            self.variables["run_modes"][run_mode_name] = variables

    def dump(self):
        import yaml

        self.logger.debug(f"writing config.yaml to '{self.file_path}")
        with open(self.file_path, "w") as fo:
            fo.write(yaml.dump(self.variables))

    def set_file_path(self, file_path):
        self.file_path = file_path

    @property
    def puck_data(self):
        return self.variables["puck_data"]

    def variable_exists(self, variable_name, variable_key):
        return variable_key in self.variables[variable_name]

    def assert_variable(self, variable_name, variable_key):
        self.assert_main_variable(variable_name)
        if not isinstance(variable_key, list):
            variable_key = [variable_key]

        for key in variable_key:
            if not self.variable_exists(variable_name, key):
                variable_name_sg = self.main_variables_pl2sg[variable_name]
                raise ConfigVariableNotFoundError(variable_name_sg, key)

    def delete_variable(self, variable_name, variable_key, **kw):
        self.assert_variable(variable_name, variable_key)

        if variable_name in self.vars_with_default and variable_key == "default":
            raise EmptyConfigVariableError(variable_name)

        variable_data = self.variables[variable_name][variable_key]
        if variable_name == "species":
            ref_name = kw["reference"]
            if ref_name in self.variables[variable_name][variable_key]:
                del self.variables[variable_name][variable_key][kw["reference"]]
            else:
                logging.warning(
                    f"reference {ref_name} is not registered under species {variable_key}"
                )

            if len(self.variables[variable_name][variable_key]) == 0:
                # this was the last reference entry.
                # Let's remove the species altogether
                del self.variables[variable_name][variable_key]
        else:
            del self.variables[variable_name][variable_key]

        return variable_data

    def process_run_mode_args(self, name=None, **kwargs):
        # typeset boolean values of run_mode
        default_run_mode = self.get_variable("run_modes", "default")

        for key, value in default_run_mode.items():
            if isinstance(value, bool) and key in kwargs.keys():
                kwargs[key] = str2bool(kwargs[key])

        return kwargs

    def process_barcode_flavor_args(self, **kw):
        bam_tags = "CR:{cell},CB:{cell},MI:{UMI},RG:{assigned}"

        # r(1|2) and then string slice
        # this is too restrictive. What about umi = "r1[12:20] + r2[:8]" ?
        # would be perfectly reasonable
        # to_match = r"r(1|2)(\[((?=-)-\d+|\d)*\:((?=-)-\d+|\d*)(\:((?=-)-\d+|\d*))*\])+$"

        UMI = kw.get("UMI", None)
        # if umi is not None and re.match(to_match, umi) is None:
        #     raise InvalidBarcodeStructureError("umi", to_match)

        cell_barcode = kw.get("cell", None)
        # if cell_barcode is not None and re.match(to_match, cell_barcode) is None:
        #     raise InvalidBarcodeStructureError("cell", to_match)

        barcode_flavor = {"bam_tags": bam_tags}

        if UMI is not None:
            barcode_flavor["UMI"] = UMI

        if cell_barcode is not None:
            barcode_flavor["cell"] = cell_barcode

        for key, value in kw.items():
            if value is not None:
                barcode_flavor[key] = value

        return barcode_flavor

    def process_adapter_args(self, name=None, seq=None):
        # print(f"processing {name} {seq}")
        if (name is None) or (seq is None):
            raise ValueError(f"need a name and a sequence: name={name} seq={seq}")

        return {name: seq}

    def process_adapter_flavor_args(self, cut_left=[], cut_right=[], name=None, **kw):
        def parse_adapters(cut):
            type_d = {
                "max_errors": float,
                "min_overlap": int,
                "min_base_qual": float,
            }
            from collections import OrderedDict

            out = []
            for c in cut:
                parts = c.split(":")
                adap_name = parts[0]
                if adap_name != "Q" and not adap_name in self.variables["adapters"]:
                    raise KeyError(
                        (
                            f"unknown adapter {adap_name}! We have "
                            f"{','.join(self.variables['adapters'])} "
                            " you can add new adapter with --add_adapter "
                        )
                    )
                d = {}
                if len(parts) > 1:
                    for param_str in parts[1:]:
                        key, value = param_str.split("=")
                        d[key] = type_d.get(key, str)(value)

                out.append({adap_name: d})

            return out

        adapter_flavor = {
            "cut_left": parse_adapters(cut_left),
            "cut_right": parse_adapters(cut_right),
        }

        for key, value in kw.items():
            if value is not None:
                adapter_flavor[key] = value

        # print(f"processed adapter_flavors {adapter_flavor}")
        return {name: adapter_flavor}

    def process_species_args(
        self,
        name=None,
        sequence=None,
        annotation=None,
        reference=None,
        STAR_index_dir=None,
        BT2_index=None,
        BT2_flags=None,
        STAR_flags=None,
    ):
        assert_file(sequence, default_value=None, extension=[".fa", ".fa.gz"])
        if annotation:
            assert_file(annotation, default_value=None, extension=[".gtf", ".gtf.gz"])

        d = dict(
            sequence=sequence,
            annotation=annotation,
        )
        # assert_file(STAR_genome, default_value=None, extension=[".fa"])
        if STAR_index_dir:
            check_star_index_compatibility(STAR_index_dir)
            d["STAR_index_dir"] = STAR_index_dir

        if BT2_index:
            d["BT2_index"] = BT2_index

        if BT2_flags:
            d["BT2_flags"] = BT2_flags

        if STAR_flags:
            d["STAR_flags"] = STAR_flags

        species_refs = self.variables["species"].get(name, {})
        species_refs[reference] = d
        self.variables["species"][name] = species_refs
        return species_refs

    def process_puck_args(
        self,
        width_um=None,
        spot_diameter_um=None,
        barcodes=None,
        coordinate_system=None,
        name=None,
    ):
        assert_file(barcodes, default_value=None, extension="all")
        assert_file(coordinate_system, default_value=None, extension="all")

        puck = {}
        if width_um is not None:
            puck["width_um"] = float(width_um)

        if spot_diameter_um is not None:
            puck["spot_diameter_um"] = float(spot_diameter_um)

        if barcodes is not None:
            puck["barcodes"] = barcodes

        if coordinate_system is not None:
            puck["coordinate_system"] = coordinate_system

        return puck

    def process_variable_args(self, variable, **kwargs):
        if variable == "barcode_flavors":
            return self.process_barcode_flavor_args(**kwargs)
        elif variable == "adapters":
            return self.process_adapter_args(**kwargs)
        elif variable == "adapter_flavors":
            return self.process_adapter_flavor_args(**kwargs)
        elif variable == "run_modes":
            return self.process_run_mode_args(**kwargs)
        elif variable == "pucks":
            return self.process_puck_args(**kwargs)
        elif variable == "species":
            return self.process_species_args(**kwargs)
        else:
            ValueError(f"Invalid variable: {variable}")

    def add_variable(self, variable, name, **kwargs):
        kwargs["name"] = name
        # It's not very clear that a cmdline arg
        # --name is absolutely REQUIRED and its value has to map somehow onto
        # an internal function name
        # @TAMAS: can you help?
        # print(f"add_variable() called with variable={variable} name={name} kw={kwargs}")

        if variable == "species":
            # for the species command, collision check is on the reference name, not the species name
            if "genome" in kwargs:
                # deprecated cmdline option --genome ... was used. Translate to
                # --sequence ... --reference=genome
                kwargs["sequence"] = kwargs["genome"]
                kwargs["reference"] = "genome"

            ref = kwargs["reference"]
            if ref in self.variables[variable].get(name, {}):
                raise DuplicateConfigVariableError(variable, f"{name}.{ref}")
            else:
                values = self.process_variable_args(variable, **kwargs)
                self.variables[variable][name] = values
        elif variable == "adapters":
            if not self.variable_exists(variable, name):
                values = self.process_variable_args(variable, **kwargs)
                self.variables[variable].update(values)
            else:
                raise DuplicateConfigVariableError(variable, name)
        else:
            if not self.variable_exists(variable, name):
                values = self.process_variable_args(variable, **kwargs)
                self.variables[variable][name] = values
            else:
                if variable in ["run_modes", "pucks", "barcode_flavors"]:
                    # drop the last s
                    variable = variable[:-1]

                raise DuplicateConfigVariableError(variable, name)

        return kwargs

    def update_variable(self, variable, name, **kwargs):
        if self.variable_exists(variable, name):
            # print(f"updating {variable} {name} {kwargs}")
            values = self.process_variable_args(variable, name=name, **kwargs)
            if variable == "adapters":
                self.variables[variable].update(values)
            elif variable == "adapter_flavors":
                raise NotImplementedError(
                    (
                        "Updating ofadapter flavors is not implemented! "
                        "please delete the flavor and then add aggain."
                    )
                )
            else:
                self.variables[variable][name].update(values)
            variable_data = self.variables[variable][name]
        else:
            if variable in ["run_modes", "pucks", "barcode_flavors"]:
                # drop the last s
                variable = variable[:-1]

            raise ConfigVariableNotFoundError(variable, name)

        return variable_data

    def get_variable(self, variable, name):
        if not self.variable_exists(variable, name):
            raise ConfigVariableNotFoundError(variable, name)
        else:
            return self.variables[variable][name]

    def get_run_mode(self, name):
        # first load the default values
        rm = RunMode(name, **self.get_variable("run_modes", name))

        default_rm = RunMode("default", **self.get_variable("run_modes", "default"))

        if rm.has_parent():
            parent_rm = self.get_run_mode(rm.parent_name)
            parent_rm.update(rm)
            rm = parent_rm

        # update with self
        default_rm.update(rm)
        rm = default_rm

        return rm

    def get_puck(self, name, return_empty=False):
        try:
            return Puck(name, **self.get_variable("pucks", name))
        except ConfigVariableNotFoundError as e:
            if not return_empty:
                raise
            else:
                return Puck(name)

    def get_barcode_flavor(self, flavor):
        return self.get_variable("barcode_flavors", flavor)

    def get_adapter_flavor(
        self, flavor, add_adapter_sequences=["cut_left", "cut_right"]
    ):
        adapter_sequences = self.variables["adapters"]
        af = self.get_variable("adapter_flavors", flavor)
        # populate each adapter clip definition in this flavor with the full sequence of the adapter
        for section, entries in af.items():
            if section in add_adapter_sequences:
                for name, d_adap in entries.items():
                    if name == "Q":
                        continue
                    d_adap["seq"] = adapter_sequences.get(
                        name,
                        f"'{name}' NOT FOUND in 'adapters:' section of config.yaml",
                    )

        return af


def get_run_mode_parser(required=True):
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        formatter_class=argparse.RawTextHelpFormatter,
        description="add/update run_mode parent parser",
        add_help=False,
    )
    parser.add_argument(
        "--name", type=str, help="name of the run_mode to be added", required=True
    )
    parser.add_argument(
        "--parent_run_mode",
        type=str,
        help="Name of the parent run_mode. All run_modes will fall back to 'default'",
    )
    parser.add_argument(
        "--umi_cutoff",
        type=int,
        nargs="+",
        help="umi_cutoff for this run_mode."
        + "the automated analysis will be run with these cutoffs",
    )
    parser.add_argument(
        "--n_beads", type=int, help="number of expected beads for this run_mode"
    )
    parser.add_argument(
        "--clean_dge",
        required=False,
        choices=bool_in_str,
        type=str,
        help="if True, the DGE will be cleaned of barcodes which overlap with primers",
    )
    parser.add_argument(
        "--detect_tissue",
        required=False,
        choices=bool_in_str,
        type=str,
        help="By default only beads having at least umi_cutoff UMI counts are analysed "
        + "during the automated analysis, all other beads are filtered out. If this "
        + "parameter is set, contiguous islands within umi_cutoff passing beads will "
        + "also be included in the analysis",
    )
    parser.add_argument(
        "--polyA_adapter_trimming",
        required=False,
        choices=bool_in_str,
        type=str,
        help="if set, reads will have polyA stretches and adapter sequence overlaps trimmed "
        + "BEFORE mapping.",
    )
    parser.add_argument(
        "--count_intronic_reads",
        required=False,
        choices=bool_in_str,
        type=str,
        help="if set, INTRONIC reads will also be countsed (apart from UTR and CDS)",
    )
    parser.add_argument(
        "--count_mm_reads",
        required=False,
        choices=bool_in_str,
        type=str,
        help="if True, multi-mappers will also be counted. For every "
        + "multimapper only reads which have one unique read mapped"
        + "to a CDS or UTR region will be counted",
    )

    parser.add_argument(
        "--mesh_data",
        required=False,
        choices=bool_in_str,
        type=str,
        help="if True, this data will be 'mehsed': a hexagonal structured "
        + "meshgrid will be created, where each new spot will have diameter"
        + " of --mesh_spot_diameter_um micron diameter and the spots will "
        + "be spaced --mesh_spot_distance_um microns apart",
    )
    parser.add_argument(
        "--mesh_type",
        required=False,
        choices=["circle", "hexagon"],
        type=str,
        help="circle: circles with diameter of --mesh_spot_diameter_um will be placed"
        + " in a hex grid, where he distance between any two circles will be "
        + "--mesh_spot_distance_um\n"
        + "hexagon: a mesh of touching hexagons will be created, with their "
        + "centers being --mesh_spot_distance_um apart. This will cover all "
        + "the data without any holes",
    )
    parser.add_argument(
        "--mesh_spot_diameter_um",
        type=float,
        required=False,
        help="diameter of mesh spot, in microns. to create a visium-style "
        + "mesh, use 55um",
    )
    parser.add_argument(
        "--mesh_spot_distance_um",
        type=float,
        required=False,
        help="distance between mesh spots in um. to create a visium-style "
        + "mesh use 100um",
    )

    return parser


def get_species_parser(required=True):
    "a parser that allows to add a reference sequence and annotation, belonging to some species"
    parser = argparse.ArgumentParser(allow_abbrev=False, add_help=False)
    parser.add_argument(
        "--reference",
        help="name of the reference (default=genome)",
        type=str,
        default="genome",
    )
    parser.add_argument("--name", help="name of the species", type=str, required=True)
    parser.add_argument(
        "--sequence",
        help="path to the sequence (.fa) file for the species/reference to be added (e.g. the genome)",
        type=str,
        required=required,
    )
    parser.add_argument(
        "--genome",
        help="[DEPRECATED] path to the genome (.fa) file for the species to be added. --genome=<arg> is a synonym for --reference=genome --sequence=<arg>",
        type=str,
        required=False,
    )

    parser.add_argument(
        "--annotation",
        help="path to the genome annotation (.gtf) file for the species to be added",
        type=str,
        default="",
        required=False,
    )
    parser.add_argument(
        "--STAR_index_dir",
        help="path to STAR index directory",
        type=str,
        required=False,
    )
    parser.add_argument(
        "--BT2_index",
        help="path to BOWTIE2 index",
        type=str,
        required=False,
    )
    parser.add_argument(
        "--BT2_flags",
        help="bt2 mapping arguments for this reference (default=mapping.smk:default_BT2_MAP_FLAGS) ",
        type=str,
        default="",
        required=False,
    )
    parser.add_argument(
        "--STAR_flags",
        help="STAR mapping arguments for this reference (default=mapping.smk:default_STAR_MAP_FLAGS)",
        type=str,
        default="",
        required=False,
    )

    return parser


def get_barcode_flavor_parser(required=True):
    parser = argparse.ArgumentParser(
        allow_abbrev=False, description="add/update barcode_flavor", add_help=False
    )
    parser.add_argument(
        "--name", help="name of the barcode flavor", type=str, required=True
    )
    parser.add_argument(
        "--umi",
        dest="UMI",
        help="structure of UMI, using python's list syntax. Example: to set UMI to "
        + "13-20 NT of Read1, use --UMI r1[12:20]. It is also possible to use the first 8nt of "
        + "Read2 as UMI: --UMI r2[0:8]",
        type=str,
        required=required,
    )
    # parser.add_argument(
    #     "--read1",
    #     help="can be used to modify or replace 'read1' sequence",
    #     type=str,
    # )

    parser.add_argument(
        "--cell-barcode",
        dest="cell",
        help="structure of CELL BARCODE, using python's list syntax. Example: to set"
        + " the cell_barcode to 1-12 nt of Read1, use --cell-barcode r1[0:12]. It is also possible "
        + " to reverse the CELL BARCODE, for instance with r1[0:12][::-1] (reversing the first 12nt of"
        + " Read1, and assigning them as CELL BARCODE).",
        type=str,
        required=required,
    )
    parser.add_argument(
        "--cell_barcode",
        dest="cell",
        help=argparse.SUPPRESS,
        type=str,
        required=False,
    )

    parser.add_argument(
        "--seq",
        help=(
            "if set, allows to modify the sequence, e.g. "
            "--seq='r2[4:]' will trim the first 4 bases. "
            "NOTE: has to be matched with --qual='r2_qual[4:]' !"
        ),
        type=str,
        default=None,
    )
    parser.add_argument(
        "--qual",
        help=(
            "if set, allows to modify the base qualities, e.g. "
            "--qual='r2_qual[4:]' will trim the first 4 quality values"
            "NOTE: has to be matched with --seq='r2[4:]' !"
        ),
        type=str,
        default=None,
    )
    parser.add_argument(
        "--min-qual-trim",
        help=(
            "Especially if you need access to bases 'at the end' of the read, "
            "that may require trimming bases that bcl2fastq already "
            "marked as non-existent (i.e. sequencing adapter). "
            "a good option for that is to clip bases from the 3' end if they "
            "have very low quality scores (default=disabled)"
        ),
        type=int,
        default=None,
    )
    parser.add_argument(
        "--min_qual_trim",
        help=argparse.SUPPRESS,
        type=int,
        default=None,
    )

    return parser


def get_adapter_parser(required=True):
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        add_help=False,
        description=("add/update adapter sequences"),
    )
    parser.add_argument("--name", help="name of the adapter", type=str, required=True)
    parser.add_argument(
        "--seq",
        help="sequence of the adapter",
        type=str,
        required=required,
    )
    return parser


def get_adapter_flavor_parser(required=True):
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        add_help=False,
        description="add/update adapter_flavor",
    )
    parser.add_argument(
        "--name",
        help=(
            "name of the adapter flavor."
            "an 'adapter_flavor' describes a recipe for trimming raw reads. "
            "this can entail clipping sequences from either end (such as TSO "
            "and/or polyA tails) as well as quality-score based trimming "
            "of bases on the reads 3' end. \n"
            "NOTE: adapters need to be first added via \n"
            "   add_adapter --name <name> --seq <sequence>\n"
            "before they can be used in an adapter_flavor. To customize the "
            "parameters for determining a match, key-value pairs can be passed "
            "to --cut_left and --cut_right, for example:\n\n"
            "   --cut_left SMART:min_overlap=10:max_errors=0.1 "
            "   --cut_right Q:min_base_qual=30 polyA:min_overlap=3:max_errors=0.25 "
        ),
        type=str,
        required=True,
    )
    parser.add_argument(
        "--cut_left",
        help=(
            "list the adapter sequences that should "
            "be trimmed from the left side of the read,"
            "e.g. --cut_left TSO_SMART SMART"
        ),
        type=str,
        nargs="+",
    )
    parser.add_argument(
        "--cut_right",
        help=(
            "list the adapter sequences that should "
            "be trimmed from the right side of the read,"
            "e.g. --cut_right Q polyG polyA "
            "NOTE: 'Q' means quality-score trimming"
        ),
        type=str,
        nargs="+",
    )
    parser.add_argument(
        "--paired_end",
        help=(
            "if consecutive reads are mates of a paired-end read,"
            "and trimming leads to a read that is shorter than "
            "min_read_length, we need to discard either both reads,"
            "or, instead of discarding, the short read can be replaced"
            " with a single 'N' [EXPERIMENTAL]"
        ),
        type=str,
        choices=["single_end", "replace_N", "discard_both"],
    )
    parser.add_argument(
        "--min_read_length",
        help=(
            "if triming reduces the length of the read below this"
            " threshold, the read should be discarded (default=18)"
        ),
        type=int,
        default=None,
    )

    return parser


def get_quant_parser(required=True):
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        add_help=False,
        description="add/update counting flavors (quant)",
    )
    parser.add_argument(
        "--name",
        help=("name of the counting flavor."),
        type=str,
        required=True,
    )
    return parser


def get_variable_action_subparsers(parent_parser, variable):

    if variable == "barcode_flavors":
        variable_singular = variable[:-1]
        variable_add_update_parser = get_barcode_flavor_parser
    elif variable == "adapters":
        variable_singular = variable[:-1]
        variable_add_update_parser = get_adapter_parser
    elif variable == "adapter_flavors":
        variable_singular = variable[:-1]
        variable_add_update_parser = get_adapter_flavor_parser
    elif variable == "pucks":
        variable_singular = variable[:-1]
        variable_add_update_parser = get_puck_parser
    elif variable == "run_modes":
        variable_singular = variable[:-1]
        variable_add_update_parser = get_run_mode_parser
    elif variable == "species":
        variable_singular = variable
        variable_add_update_parser = get_species_parser
    elif variable == "quant":
        variable_singular = variable
        variable_add_update_parser = get_quant_parser

    else:
        print(f"suspicious variable {variable}")

    command_help = {
        "list": f"list {variable} and their settings",
        "delete": f"delete {variable_singular}",
        "add": f"add a new {variable_singular}",
        "update": f"update an existing {variable_singular}",
    }

    # list command
    list_parser = parent_parser.add_parser(
        f"list-{variable.replace('_', '-')}",
        description=command_help["list"],
        help=command_help["list"],
    )
    list_parser.set_defaults(func=list_variables_cmdline, variable=variable)
    # snake_case for backward compatibility. remove in a future update
    list_parser_legacy = parent_parser.add_parser(
        f"list_{variable}",
    )
    list_parser_legacy.set_defaults(func=list_variables_cmdline, variable=variable)

    func = add_update_delete_variable_cmdline

    # delete command
    delete_parser = parent_parser.add_parser(
        f"delete-{variable_singular.replace('_', '-')}",
        description=command_help["delete"],
        help=command_help["delete"],
    )
    delete_parser_legacy = parent_parser.add_parser(
        f"delete_{variable_singular}",
    )
    delete_parser.add_argument(
        "--name",
        help=f"name of the {variable_singular} to be deleted",
        type=str,
        required=True,
    )
    delete_parser_legacy.add_argument(
        "--name",
        help=f"name of the {variable_singular} to be deleted",
        type=str,
        required=True,
    )
    if variable == "species":
        delete_parser.add_argument(
            "--reference",
            help=f"name of the reference to be deleted (genome, rRNA, ...)",
            type=str,
            required=True,
        )
        delete_parser_legacy.add_argument(
            "--reference",
            help=f"name of the reference to be deleted (genome, rRNA, ...)",
            type=str,
            required=True,
        )
    delete_parser.set_defaults(func=func, action="delete", variable=variable)
    delete_parser_legacy.set_defaults(func=func, action="delete", variable=variable)

    # add command
    add_parser = parent_parser.add_parser(
        f"add-{variable_singular.replace('_', '-')}",
        parents=[variable_add_update_parser()],
        description=command_help["add"],
        help=command_help["add"],
    )
    add_parser_legacy = parent_parser.add_parser(
        f"add_{variable_singular}",
        parents=[variable_add_update_parser()],
    )
    add_parser.set_defaults(func=func, action="add", variable=variable)
    add_parser_legacy.set_defaults(func=func, action="add", variable=variable)

    # update command
    update_parser = parent_parser.add_parser(
        f"update-{variable_singular.replace('_', '-')}",
        parents=[variable_add_update_parser(False)],
        description=command_help["update"],
        help=command_help["update"],
    )
    update_parser_legacy = parent_parser.add_parser(
        f"update_{variable_singular}",
        parents=[variable_add_update_parser(False)],
    )
    update_parser.set_defaults(func=func, action="update", variable=variable)
    update_parser_legacy.set_defaults(func=func, action="update", variable=variable)


def setup_config_parser(parent_parser_subparsers):
    parser_config = parent_parser_subparsers.add_parser(
        "config", help="configure spacemake"
    )
    parser_config_subparsers = parser_config.add_subparsers(
        help="config sub-command help"
    )

    for variable in ConfigFile.main_variables_pl2sg.keys():
        get_variable_action_subparsers(parser_config_subparsers, variable)

    return parser_config


@message_aggregation(logger_name)
def add_update_delete_variable_cmdline(args):
    # set the name and delete from dictionary
    name = args["name"]
    variable = args["variable"]
    action = args["action"]

    import yaml

    config = get_global_config()
    config.assert_main_variable(variable)

    # remove the args from the dict
    del args["action"]
    del args["variable"]
    del args["name"]

    if action == "add":
        func = config.add_variable
    elif action == "update":
        func = config.update_variable
    elif action == "delete":
        func = config.delete_variable

    var_variables = func(variable, name, **args)
    # print and dump config file
    config.logger.info(yaml.dump(var_variables, sort_keys=False))
    config.dump()


@message_aggregation(logger_name)
def list_variables_cmdline(args):
    variable = args["variable"]
    del args["variable"]
    import yaml

    config = get_global_config()

    config.logger.info(f"Listing {variable}")
    config.logger.info(yaml.dump(config.variables[variable]))



import mrfifo as mf
import logging


def parse_args():
    from spacemake.util import make_minimal_parser

    parser = make_minimal_parser("BamTagHistogram")

    parser.add_argument("--parallel", type=int, default=8)
    parser.add_argument("--input", default="/dev/stdin")
    parser.add_argument("--output", default="/dev/stdout")
    parser.add_argument(
        "--prefix-size",
        default=4,
        type=int,
        help=(
            "how many letters of the tag value are used to split the stream. "
            "default=4 allows for up to (alphabet_size)^4 distinct parallel workers. "
            "will be spread across workers by mod <args.parallel>"
        ),
    )
    parser.add_argument("--prefix-alphabet", default="ACGTN")
    parser.add_argument("--min-count", default=10, type=int)
    parser.add_argument(
        "--sort-mem",
        default=8,
        type=int,
        help="how many GB are allowed to be used for sorting (default=8)",
    )
    parser.add_argument(
        "--tag", default="CB", help="which BAM tag to count (default='CB')"
    )

    return parser.parse_args()


def CB_distributor(
    input, outputs, tag="CB", prefix_size=3, prefix_alphabet="ACGTN", n=8, **kw
):
    "ensure that the FIFOs are not managed"
    assert type(input) is str
    logger = logging.getLogger("mrfifo.parts.CB_distributor")
    logger.debug(
        f"reading from {input}, writing to {outputs} "
        f"tag={tag} prefix_size={prefix_size} prefix_alphabet={prefix_alphabet} "
        f"kw={kw}"
    )

    lkup = {}
    from itertools import product

    i = 0
    for letters in product(*([prefix_alphabet] * prefix_size)):
        prefix = "".join(letters).encode("ascii")
        lkup[prefix] = i % n
        i += 1

    # for k, v in sorted(lkup.items()):
    #     print(f"{k}\t{v}")

    from mrfifo.fast_loops import distribute_by_substr

    tag_lead = b"\t" + tag.encode("ascii") + b":Z:"
    logger.debug(
        f"scanning for tag-lead {tag_lead} and using next {prefix_size} bytes as prefix"
    )
    res = distribute_by_substr(
        fin_name=input,
        fifo_names=outputs,
        sub_lookup=lkup,
        sub_size=prefix_size,
        sub_lead=tag_lead,
        # **kw,
    )
    logger.debug("distribution complete")
    return res


def tag_counter(input, output, tag="CB", min_count=10, **kw):
    from collections import defaultdict

    counter = defaultdict(int)
    stats = defaultdict(int)
    import re

    pattern = re.compile(f"{tag}:Z:(\S+)")
    for sam_line in input:
        stats["n_records"] += 1
        flags = int(sam_line.split("\t")[1])
        if flags & 256:
            # 'not primary alignment' bit is set
            stats["n_secondary"] += 1
            continue

        if m := re.search(pattern, sam_line):
            stats["n_tagged"] += 1
            tag_val = m.groups(0)[0]
            counter[tag_val] += 1

    stats["n_values"] = len(counter)
    for value, count in counter.items():
        if count >= min_count:
            stats["n_above_cut"] += 1
            output.write(f"{count}\t{value}\n")

    return stats


def sort_function(input, output, n=8, sort_mem_gigs=8, header=None):
    import os

    if header is None:
        header = rf"# INPUT={args.input} TAG={args.tag} FILTER_PCR_DUPLICATES=false READ_QUALITY=0\n"

    if output.endswith(".gz"):
        cmd = (
            f'{{ printf "{header}"; sort -rnk 1 -S {sort_mem_gigs}G --parallel={n} {input}; }}'
            f"| python -m isal.igzip -c > {output}"
        )
    else:
        cmd = f'{{ printf "{header}"; sort -rnk 1 -S {sort_mem_gigs}G --parallel={n} {input}; }} > {output}'

    import subprocess

    subprocess.call(cmd, shell=True)


def main(args):
    w = (
        mf.Workflow("BamTagHistogram", total_pipe_buffer_MB=4)
        .BAM_reader(
            input=args.input,
            mode="S",
            threads=4,
        )
        .distribute(
            input=mf.FIFO("input_sam", "rt"),
            outputs=mf.FIFO("dist_{n}", "wt", n=args.parallel),
            func=CB_distributor,
            tag=args.tag,
            prefix_size=args.prefix_size,
            prefix_alphabet=args.prefix_alphabet,
            n=args.parallel,
        )
        .workers(
            func=tag_counter,
            tag=args.tag,
            input=mf.FIFO("dist_{n}", "rt"),
            output=mf.FIFO("counts_{n}", "wt"),
            n=args.parallel,
            min_count=args.min_count,
        )
        .collect(
            inputs=mf.FIFO("counts_{n}", "rt", n=args.parallel),
            output=mf.FIFO("unsorted", "wt"),
            chunk_size=1,
        )
        .funnel(
            input=mf.FIFO("unsorted", "rt"),
            output=args.output,
            func=sort_function,
            _manage_fifos=False,
        )
        .run()
    )
    stats = mf.util.CountDict()
    for jobname, d in w.result_dict.items():
        if "worker" in jobname:
            stats.add_other_stats(d)

    df = stats.get_stats_df()
    df["input"] = args.input
    print(df.set_index("input"))
    return w


if __name__ == "__main__":
    args = parse_args()
    import spacemake.util as util

    util.setup_logging(args)
    main(args)



import re

def parse_ribo_log(ribo_log_file):
    # before the log, there can be some perl warnings prepended. so we need to find the
    # first line outputed by bwa mem
    input_reads = 0
    aligned_reads = 0

    # ribo log summary line: first line of the summary
    first_line_regex = r'^\d+ reads; of these:$'
    first_line_found = False

    line_n = 0

    with open(ribo_log_file) as f:
        for line in f:
            stripped_line = line.strip()
            
            if stripped_line == 'no_rRNA_index':
                input_reads = -1
                aligned_reads = -1
                break

            if not first_line_found:
                if re.match(first_line_regex, stripped_line) is not None:
                    first_line_found = True
                    line_n = 0
                else:
                    # keep looking for first line
                    continue

            if line_n == 0:
                input_reads = input_reads + int(stripped_line.split(' ')[0])
            elif line_n == 3 or line_n == 4:
                aligned_reads = aligned_reads + int(stripped_line.split(' ')[0])
            # reset after the fifth line, this is needed if there are several ribolog files
            # appended one after the other. this is the case for merged samples
            elif line_n == 5:
                first_line_found = False

            line_n = line_n + 1
            
    
    if input_reads <= 0:
        return (None, None)
    else:
        return (aligned_reads, input_reads)


if snakemake.params.ribo_log == "no_rRNA_index":
    input_reads = -1
    aligned_reads = -1

else:
    aligned_reads, input_reads = parse_ribo_log(snakemake.params.ribo_log)

with open(snakemake.output[0], 'w') as fo:
    fo.write(f'aligned_reads\t{aligned_reads}\n')
    fo.write(f'input_reads\t{input_reads}\n')



import pytest
import sys
import os
from spacemake.cmdline import *

from fixtures import (
    initialized_root,
    with_species,
    with_tile_test_data,
    configured_root,
    sm,
    spacemake_dir,
)


def test_init(initialized_root):
    assert os.path.exists(initialized_root.as_posix() + "/config.yaml")


def test_parsers(configured_root):
    get_project_sample_parser()
    get_add_sample_sheet_parser()

    parser = get_sample_main_variables_parser()
    get_action_sample_parser(parser.add_subparsers(), "add", lambda *argc, **kw: None)

    get_sample_extra_info_parser()
    get_data_parser()
    get_run_parser()

    make_main_parser()


def test_subcommands(configured_root):
    # test everything after init
    os.chdir(configured_root.as_posix())

    sm("projects", "list")
    sm("config", "list_adapter_flavors")


def test_config_barcodes(initialized_root):
    os.chdir(initialized_root.as_posix())

    # add
    sm(
        "config",
        "add-barcode-flavor",
        "--name",
        "fc_sts_miniseq",
        "--umi",
        "r2[0:9]",
        "--cell-barcode",
        "r1[2:27]",
    )
    # add same bc flavor twice? -> error
    sm(
        "config",
        "add-barcode-flavor",
        "--name",
        "fc_sts_miniseq",
        "--umi",
        "r2[0:9]",
        "--cell-barcode",
        "r1[2:27]",
        expect_fail=True,
    )
    # delete
    sm(
        "config",
        "delete-barcode-flavor",
        "--name",
        "fc_sts_miniseq",
    )
    # re-add
    sm(
        "config",
        "add-barcode-flavor",
        "--name",
        "fc_sts_miniseq",
        "--umi",
        "r2[0:8]",
        "--cell-barcode",
        "r1[2:27]",
    )
    # update
    sm(
        "config",
        "update-barcode-flavor",
        "--name",
        "fc_sts_miniseq",
        "--umi",
        "r2[0:9]",
    )


def test_config_adapters(initialized_root):
    os.chdir(initialized_root.as_posix())
    # add
    sm(
        "config",
        "add-adapter",
        "--name",
        "testy",
        "--seq",
        "ACGTACGTACGTACGT",
    )
    sm(
        "config",
        "add-adapter",
        "--name",
        "testy",
        "--seq",
        "ACGTACGTACGTACGT",
        expect_fail=True,
    )
    # update
    sm(
        "config",
        "update-adapter",
        "--name",
        "testy",
        "--seq",
        "ACGTACGTACGTACGTA",
    )
    # delete
    sm("config", "delete-adapter", "--name", "testy")
    # re-add
    sm(
        "config",
        "add-adapter",
        "--name",
        "testy",
        "--seq",
        "ACGTACGTACGTACGTACGT",
    )


def test_config_adapter_flavors(initialized_root):
    os.chdir(initialized_root.as_posix())

    test_flavor = (
        "--name",
        "testy1",
        "--cut_left",
        "SMART:min_overlap=10:max_errors=0.1",
        "--cut_right",
        "Q:min_base_qual=30",
        "polyG:min_overlap=3:max_errors=0.2",
        "polyA:min_overlap=3:max_errors=0.25",
    )

    # add
    sm("config", "add-adapter-flavor", *test_flavor)
    sm("config", "add-adapter-flavor", *test_flavor, expect_fail=True)
    # update -> not implemented
    sm("config", "update-adapter-flavor", "--name", "testy1", expect_fail=True)
    # delete
    sm("config", "delete-adapter-flavor", "--name", "testy1")
    # re-add
    sm("config", "add-adapter-flavor", *test_flavor)


def test_species(with_species):
    pass  # It's all in the fixture `with_species`


def test_puck(initialized_root):
    os.chdir(initialized_root.as_posix())
    test_puck = (
        "--name=test_puck",
        "--width_um=2",
        "--spot_diameter_um=0.5",
    )
    # add
    sm("config", "add_puck", *test_puck)
    sm("config", "add_puck", *test_puck, expect_fail=True)
    # update
    sm("config", "update_puck", "--name=test_puck", "--width_um=3")
    # delete
    sm("config", "delete_puck", "--name=test_puck")
    # re-add
    sm("config", "add_puck", *test_puck)


def test_runmode(initialized_root):
    os.chdir(initialized_root.as_posix())
    # add
    sm("config", "add_run_mode", "--name=spatial_rm", "--umi_cutoff=10")
    sm("config", "add_run_mode", "--name=spatial_rm2", "--umi_cutoff=10")
    # delete
    sm("config", "delete_run_mode", "--name=spatial_rm2")
    # edit
    sm("config", "update_run_mode", "--name=spatial_rm", "--umi_cutoff=1")
    from spacemake.config import get_global_config

    config = get_global_config()
    config.dump()


def test_issue_88(initialized_root):
    os.chdir(initialized_root.as_posix())
    # name added to config.yaml when creating a new run mode #88
    # add
    sm("config", "add_run_mode", "--name", "newrunmode")
    from spacemake.config import get_global_config

    config = get_global_config()
    config.dump()
    inserted = config.variables["run_modes"]["newrunmode"]
    print("here's what we inserted", inserted)
    assert "name" not in inserted


def test_sample(configured_root):
    os.chdir(configured_root.as_posix())
    test_sample = (
        f"--R1={spacemake_dir}/test_data/simple.reads1.fastq.gz",
        f"--R2={spacemake_dir}/test_data/simple.reads2.fastq.gz",
        "--map-strategy=rRNA:bowtie2->genome:STAR:final",
        "--species=test_hsa",
    )

    # add
    sm("projects", "add-sample", "--project-id=test", "--sample-id=test1", *test_sample)
    # delete
    sm(
        "projects",
        "delete-sample",
        "--project-id=test",
        "--sample-id=test1",
    )
    # re-add
    sm("projects", "add-sample", "--project-id=test", "--sample-id=test1", *test_sample)
    # update
    sm(
        "projects",
        "update-sample",
        "--project-id=test",
        "--sample-id=test1",
        "--map-strategy=rRNA:bowtie2->miRNA:bowtie2->genome:STAR",
    )
    # add a second run mode to the sample
    sm(
        "projects",
        "update-sample",
        "--project-id=test",
        "--sample-id=test1",
        "--run-mode",
        "openst",
        "default",
    )


def test_fill_project_df(with_species):
    os.chdir(with_species.as_posix())
    test_project_data = [
        (
            "test_hsa",
            "test",
            "test_01",
            f"{spacemake_dir}/test_data/simple.reads1.fastq.gz",
            f"{spacemake_dir}/test_data/simple.reads2.fastq.gz",
            "--map-strategy=genome:STAR:final",
        ),
        (
            "test_hsa",
            "test",
            "test_02.2",
            f"{spacemake_dir}/test_data/simple.reads1.fastq.gz",
            f"{spacemake_dir}/test_data/simple.reads2.fastq.gz",
            "--map-strategy=rRNA:bowtie2->miRNA:bowtie2->genome:STAR:final",
        ),
        (
            "test_hsa",
            "test",
            "test_03_nofinal",
            f"{spacemake_dir}/test_data/simple.reads1.fastq.gz",
            f"{spacemake_dir}/test_data/simple.reads2.fastq.gz",
            "--map-strategy=rRNA:bowtie2->miRNA:bowtie2->genome:STAR",
        ),
        (
            "test_hsa",
            "test",
            "test_bulk",
            "None",
            f"{spacemake_dir}/test_data/simple.reads2.fastq.gz",
            "--map-strategy=rRNA:bowtie2->miRNA:bowtie2->genome:STAR:final"
            " --barcode_flavor=visium",
        ),
        (
            "test_hsa",
            "tile",
            "tile_1",
            f"{spacemake_dir}/test_data/simple.reads1.fastq.gz",
            f"{spacemake_dir}/test_data/simple.reads2.fastq.gz",
            (
                "--map-strategy=rRNA:bowtie2->miRNA:bowtie2->genome:STAR:final"
                f" --puck-barcode-file {spacemake_dir}/test_data/tile_1.txt"
                " --puck slide_seq --run-mode slide_seq"
            ),
        ),
        (
            "test_hsa",
            "tile",
            "tile_2",
            f"{spacemake_dir}/test_data/simple.reads1.fastq.gz",
            f"{spacemake_dir}/test_data/simple.reads2.fastq.gz",
            (
                "--map-strategy=rRNA:bowtie2->miRNA:bowtie2->genome:STAR:final"
                f" --puck-barcode-file {spacemake_dir}/test_data/tile_2.txt"
                " --puck slide_seq --run-mode slide_seq"
            ),
        ),
        (
            "test_hsa",
            "tile",
            "tile_both",
            f"{spacemake_dir}/test_data/simple.reads1.fastq.gz",
            f"{spacemake_dir}/test_data/simple.reads2.fastq.gz",
            (
                "--map-strategy=rRNA:bowtie2->miRNA:bowtie2->genome:STAR:final"
                f" --puck-barcode-file {spacemake_dir}/test_data/tile_1.txt {spacemake_dir}/test_data/tile_2.txt"
                " --puck slide_seq --run-mode slide_seq"
            ),
        ),
    ]
    for species, project_id, sample_id, R1, R2, extra in test_project_data:
        # add
        sm(
            "projects",
            "add-sample",
            f"--project-id={project_id}",
            f"--sample-id={sample_id}",
            f"--species={species}",
            f"--R1={R1}",
            f"--R2={R2}",
            *extra.split(" "),
        )


def test_run(configured_root, dry=True):
    # test everything after init
    os.chdir(configured_root.as_posix())
    if dry:
        sm("run", "-np", "--cores=8")
    else:
        sm("run", "-p", "--cores=8")


@pytest.mark.big_download
def test_tiles(with_tile_test_data):
    os.chdir(with_tile_test_data.as_posix())
    sm(
        "config",
        "add-puck",
        "--name=openst_fc_010",
        "--coordinate_system=fc_010_coordinate_system.csv",
        "--width_um=1200",
        "--spot_diameter_um=0.6",
    )
    sm(
        "projects",
        "add-sample",
        "--project-id=fc_sts_75",
        "--sample-id=fc_sts_75_1b",
        "--R1=fc_sts_75_1b_S1_R1_001_subsampled.fastq.gz",
        "--R2=fc_sts_75_1b_S1_R2_001_subsampled.fastq.gz",
        "--species=mouse",
        "--puck=openst_fc_010",
        "--run-mode=openst",
        "--barcode-flavor=openst",
        "--puck-barcode-file",
        "fc_010_L2_tile_2558.txt.gz",
        "fc_010_L2_tile_2559.txt.gz",
        "fc_010_L4_tile_2505.txt.gz",
        "--map-strategy=phiX:bowtie2->rRNA:bowtie2->genome:STAR",
    )


def test_tiles_nomatch(with_species, dry=False):
    os.chdir(with_species.as_posix())
    sm(
        "config",
        "add-puck",
        "--name=openst_min",
        f"--coordinate_system={spacemake_dir}/test_data/test_coordinate_system.csv",
        "--width_um=1200",
        "--spot_diameter_um=0.6",
    )
    sm(
        "projects",
        "add-sample",
        "--project-id=fc_test",
        "--sample-id=fc_test_nomatch",
        f"--R1={spacemake_dir}/test_data/simple.reads1.fastq.gz",
        f"--R2={spacemake_dir}/test_data/simple.reads2.fastq.gz",
        "--puck=openst_min",
        "--run-mode=openst",
        "--barcode-flavor=openst",
        "--puck-barcode-file",
        f"{spacemake_dir}/test_data/tile_1.txt",
        f"{spacemake_dir}/test_data/tile_2.txt",
        f"{spacemake_dir}/test_data/tile_3.txt",
        f"{spacemake_dir}/test_data/tile_4.txt",
        "--map-strategy=rRNA:bowtie2->genome:STAR:final",
        "--species=test_hsa",
    )
    if dry:
        sm("run", "-np", "--cores=8")
    else:
        sm("run", "-p", "--cores=8")
        assert os.path.exists(
            "projects/fc_test/processed_data/fc_test_nomatch/illumina/complete_data/dge/dge.all.polyA_adapter_trimmed.mm_included.100000_beads_no_spatial_data.h5ad"
        )
        assert not os.path.exists(
            "projects/fc_test/processed_data/fc_test_nomatch/illumina/complete_data/dge/dge.all.polyA_adapter_trimmed.mm_included.spatial_beads.mesh_7_hexagon_puck_collection.h5ad"
        )


def test_puck_collection(with_species, dry=False):
    sm(
        "config",
        "add-puck",
        "--name=openst_min",
        f"--coordinate_system={spacemake_dir}/test_data/test_coordinate_system.csv",
        "--width_um=1200",
        "--spot_diameter_um=0.6",
    )
    sm(
        "projects",
        "add-sample",
        "--project-id=fc_test",
        "--sample-id=fc_test_collection",
        f"--R1={spacemake_dir}/test_data/simple.reads1.fastq.gz",
        f"--R2={spacemake_dir}/test_data/simple.reads2.fastq.gz",
        "--puck=openst_min",
        "--run-mode=openst",
        "--barcode-flavor=openst",
        "--puck-barcode-file",
        f"{spacemake_dir}/test_data/tile_1.txt",
        f"{spacemake_dir}/test_data/tile_2.txt",
        f"{spacemake_dir}/test_data/tile_3.txt",
        f"{spacemake_dir}/test_data/tile_4.txt",
        f"{spacemake_dir}/test_data/tile_5.txt",
        f"{spacemake_dir}/test_data/tile_6.txt",
        "--map-strategy=rRNA:bowtie2->genome:STAR:final",
        "--species=test_hsa",
    )
    if dry:
        sm("run", "-np", "--cores=8")
    else:
        sm("run", "-p", "--cores=8")
        assert os.path.exists(
            "projects/fc_test/processed_data/fc_test_collection/illumina/complete_data/dge/dge.all.polyA_adapter_trimmed.mm_included.spatial_beads.mesh_7_hexagon_puck_collection.h5ad"
        )


def test_puck_single_tile(with_species, dry=False):
    sm(
        "config",
        "add-puck",
        "--name=openst_min",
        f"--coordinate_system={spacemake_dir}/test_data/test_coordinate_system.csv",
        "--width_um=1200",
        "--spot_diameter_um=0.6",
    )
    sm(
        "projects",
        "add-sample",
        "--project-id=fc_test",
        "--sample-id=fc_test_single_tile",
        f"--R1={spacemake_dir}/test_data/simple.reads1.fastq.gz",
        f"--R2={spacemake_dir}/test_data/simple.reads2.fastq.gz",
        "--puck=openst_min",
        "--run-mode=openst",
        "--barcode-flavor=openst",
        "--puck-barcode-file",
        f"{spacemake_dir}/test_data/tile_5.txt",
        "--map-strategy=rRNA:bowtie2->genome:STAR:final",
        "--species=test_hsa",
    )
    if dry:
        sm("run", "-np", "--cores=8")
    else:
        sm("run", "-p", "--cores=8")
        assert os.path.exists(
            "projects/fc_test/processed_data/fc_test_single_tile/illumina/complete_data/dge/dge.all.polyA_adapter_trimmed.mm_included.spatial_beads.mesh_7_hexagon_puck_collection.h5ad"
        )


def test_merge(with_species):
    os.chdir(with_species.as_posix())
    test_samples = [
        (
            "test",
            "test_01",
            f"--R1={spacemake_dir}/test_data/simple.reads1.fastq.gz",
            f"--R2={spacemake_dir}/test_data/simple.reads2.fastq.gz",
            "--map-strategy=rRNA:bowtie2->genome:STAR:final",
            "--species=test_hsa",
        ),
        (
            "test",
            "test_01b",
            f"--R1={spacemake_dir}/test_data/simple.reads1.fastq.gz",
            f"--R2={spacemake_dir}/test_data/simple.reads2.fastq.gz",
            "--map-strategy=rRNA:bowtie2->genome:STAR:final",
            "--species=test_hsa",
        ),
        (
            "test",
            "test_02",
            f"--R1={spacemake_dir}/test_data/simple.reads1.fastq.gz",
            f"--R2={spacemake_dir}/test_data/simple.reads2.fastq.gz",
            "--map-strategy=genome:STAR:final",
            "--species=test_hsa",
        ),
    ]

    # add
    for sample in test_samples:
        sm(
            "projects",
            "add-sample",
            f"--project-id={sample[0]}",
            f"--sample-id={sample[1]}",
            *sample[2:],
        )

    sm(
        "projects",
        "merge-samples",
        "--merged-project-id=test",
        "--merged-sample-id=test_merged",
        "--sample-id-list",
        "test_01",
        "test_02",
        expect_fail=True,  # different map_strategies
    )
    sm(
        "projects",
        "merge-samples",
        "--merged-project-id=test",
        "--merged-sample-id=test_merged",
        "--sample-id-list",
        "test_01",
        "test_01b",
    )
    sm("projects", "delete-sample", f"--project-id=test", f"--sample-id=test_02")
    sm("run", "-p", "--cores=8")


def test_downsample(with_species):
    sm(
        "projects",
        "add-sample",
        f"--project-id=test",
        f"--sample-id=test_01",
        f"--R1={spacemake_dir}/test_data/simple.reads1.fastq.gz",
        f"--R2={spacemake_dir}/test_data/simple.reads2.fastq.gz",
        "--map-strategy=genome:STAR:final",
        "--species=test_hsa",
    )
    sm("run", "downsample", "--sample-id-list", "test_01", "--cores=8", "-p")


def test_from_scratch(with_species):
    test_sample(with_species)
    test_run(with_species, dry=False)



import importlib.metadata

__version__ = "0.9.1b"
__author__ = [
    "Nikos Karaiskos",
    "Tamas Ryszard Sztanka-Toth",
    "Marvin Jens",
    "Daniel Leon-Perinan",
]
__license__ = "GPL"
__email__ = [
    "nikolaos.karaiskos@mdc-berlin.de",
    "tamasryszard.sztanka-toth@mdc-berlin.de",
    "marvin.jens@charite.de",
    "daniel.leonperinan@mdc-berlin.de",
]

author_contributions = """
Spacemake is built on snakemake scripts originally developed by Nikos Karaiskos
for the analysis of dropseq data. These gradually evolved into a robust workflow for
spatial transcriptomics data analysis that was improved and generalized to work
with different ST technologies by Tamas Ryszard Sztanka-Toth. Marvin Jens contributed
longread analysis code and support for converting fastq to BAM as a first step.
Many features of the automated analysis and integration with Novosparc were added by
Tamas, in close collaboration with Nikos, culminating in the first spacemake
publication:

    https://doi.org/10.1093/gigascience/giac064

Marvin then added new building blocks to successively replace the java-based 
dropseq tools with python/pysam based code: cutadapt_bam.py, annotator.py, as well
as the ability to align raw reads to multiple indices, in close collaboration
with Nikos & Tamas.

Spacemake is actively maintained by Dani, Marvin and Nikos.
"""

roadmap = [
    (
        "0.5.5",
        "universal ST support and utility, novosparc integration. Sztanka-Toth et al. 2022",
    ),
    (
        "0.7",
        "support multiple mapping indices, bulk samples, custom user-defined snakemake rules",
    ),
    ("0.7.7", "much lower RAM usage and speed boost"),
    (
        "0.8",
        "much faster cmdline. 10x speed-up for fastq_to_uBAM. Lower memory for BamTagHistogram replacement",
    ),
    (
        "0.9",
        "CRAM on-disk format together with other tweaks -> disk footprint down to 1/3 of 0.8. Replace pre-processing with cutadapt code",
    ),
    (
        "1.x",
        "completely replace dropseq tools. Own annotator and towards entirely scanpy workflow",
    ),
    ("1.x", "efficient handling of 1E8+ spatial barcodes (seq-scope etc.)"),
    ("1.x", "add interactive data exploration support (shiny?)"),
    ("2.x", "cmdline interface cleanup and remote API support"),
    ("2.x", "cython magic to speed up parallel BAM processing via shared memory"),
]



import multiprocessing as mp

import pytest
import sys
import os
from spacemake.bin.fastq_to_uBAM import *


spacemake_dir = os.path.dirname(__file__) + "/../"


@pytest.fixture(scope="session")
def test_root(tmp_path_factory):
    tmp = tmp_path_factory.mktemp("root")
    sm_path = os.path.dirname(__file__)
    # make a tmp-copy of the test_config.yaml
    # def_config = os.path.join(sm_path, "../test_data/test_config.yaml")
    # os.system(f"cp {def_config} {tmp / 'config.yaml'}")

    # test_pdf =  os.path.join(sm_path, "../test_data/test_project_df.csv")
    # os.system(f"cp {test_pdf} {tmp / 'project_df.csv'}")

    return tmp


def sm(*argc, expect_fail=False):
    sys.argv = [
        "fastq_to_uBAM.py",
    ] + list(argc)
    res = cmdline()
    print("got result", res)
    from spacemake.errors import SpacemakeError

    if expect_fail:
        assert isinstance(res, SpacemakeError) == True
    else:
        assert isinstance(res, Exception) == False


def test_help():
    try:
        sm("--help")
    except SystemExit:
        pass


def test_dropseq():
    sm(
        "--read1",
        spacemake_dir + "test_data/reads_chr22_R1.fastq.gz",
        "--read2",
        spacemake_dir + "test_data/reads_chr22_R2.fastq.gz",
        "--out-file",
        "/dev/null",
    )


def test_single():
    sm(
        "--read2",
        spacemake_dir + "test_data/reads_chr22_R2.fastq.gz",
        "--out-file",
        "/dev/null",
        """--cell='"A"'""",
    )


def test_preprocessing():
    pre = PreProcessor("quality:left=25,right=25")

    qual = "IIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII"
    sdata = SeqData(
        "@NS_mRNA:7 1:N:0:ACTGAGCG",
        "CCTGCTGGGAGGGG",
        "IIIIIIIIIIIIII",
        "CCTGCTGGGAGGGGGTGGGGGGAGGAGGAAGAGGTGGGGCTCTACTCTGATTAATTA",
        qual,
    )
    sdata = pre.process(sdata)
    assert sdata.q2 == qual
    assert len(sdata.tags) == 0

    qual = "###IIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII###I#I#"
    sdata.q2 = qual
    res = pre.process(sdata)
    # print(res)
    tag_d = dict(res.tags)
    assert tag_d["T5"] == ["3"]
    assert tag_d["T3"] == ["7"]

    pre = PreProcessor("quality:left=25,right=25;polyA")

    sdata = SeqData(
        "@NS_mRNA:7 1:N:0:ACTGAGCG",
        "CCTGCTGGGAGGGG",
        "IIIIIIIIIIIIII",
        "CCTGCTGGGAGGGGGTGGGGGGAGGAGGAAGAGGTGGGGCTCTACTCTGATTAATTAAAAAAAAAGAGAAAAAAAAAAAAAAAAGGG",
        "IIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIEEIEEIE#IIIIIIIIIIIIIIEIII###",
    )

    res = pre.process(sdata)
    tag_d = dict(res.tags)
    assert tag_d["T3"] == ["3", "28"]


if __name__ == "__main__":
    test_preprocessing()

# def test_minqual():
#     sm(
#         "--read2",
#         spacemake_dir + "test_data/reads_chr22_R2.fastq.gz",
#         "--out-bam",
#         "/dev/null",
#         "--min-qual",
#         "30",
#         """--cell='"A"'""",
#     )


def test_issue135():
    import spacemake.bin.fastq_to_uBAM as ubam
    from argparse import Namespace
    import io

    f1 = io.StringIO(
        "\n".join(
            [
                "@QNAME MUST NOT HAVE WHITESPACE",
                "CCTGCTGGGAGGGG",
                "+",
                "IIIIIIIIIIIIII",
                "",
            ]
        )
    )

    f2 = io.StringIO(
        "\n".join(
            [
                "@QNAME MUST NOT HAVE WHITESPACE",
                "CCTGCTGGGAGGGGGTGGGGGGAGGAGGAAGAGGTGGGGCTCTACTCTGATTAATTA",
                "+",
                "IIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII",
                "",
            ]
        )
    )

    pre = ubam.PreProcessor("quality:left=25,right=25;barcode")

    for sdata in ubam.SeqData.from_paired_end(f1, f2):
        sdata = pre.process(sdata)

        sam = sdata.render_SAM(flag=4)
        cols = sam.split()
        assert cols[0] == "QNAME"
        assert cols[1] == "4"
    # print(sam)


if __name__ == "__main__":
    test_issue135()
